import numpy as np

from stimupy.components import waves
from stimupy.components.shapes import disc, rectangle
from stimupy.stimuli import place_targets

__all__ = [
    "sine_linear",
    "square_linear",
    "staircase_linear",
    "sine_radial",
    "square_radial",
    "staircase_radial",
    "sine_rectilinear",
    "square_rectilinear",
    "staircase_rectilinear",
    "sine_angular",
    "square_angular",
    "staircase_angular",
]


def sine_linear(
    visual_size=None,
    ppd=None,
    shape=None,
    frequency=None,
    n_bars=None,
    bar_width=None,
    period="ignore",
    rotation=0.0,
    phase_shift=0,
    intensities=(0.0, 1.0),
    target_indices=(),
    intensity_target=0.5,
    origin="corner",
    round_phase_width=True,
):
    """Linear (horizontal, vertical, oblique) sine-wave grating, with some phase(s) as target(s)

    Parameters
    ----------
    visual_size : Sequence[Number, Number], Number, or None (default)
        visual size [height, width] of image, in degrees
    ppd : Sequence[Number, Number], Number, or None (default)
        pixels per degree [vertical, horizontal]
    shape : Sequence[Number, Number], Number, or None (default)
        shape [height, width] of image, in pixels
    frequency : Number, or None (default)
        spatial frequency of grating, in cycles per degree visual angle
    n_bars : int, or None (default)
        number of bars in the grating
    bar_width : Number, or None (default)
        width of a single bar, in degrees visual angle
    period : "even", "odd", "either" or "ignore" (default)
        ensure whether the grating has "even" number of phases, "odd"
        number of phases, either or whether not to round the number of
        phases ("ignore")
    rotation : float, optional
        rotation (in degrees), counterclockwise, by default 0.0 (horizontal)
    phase_shift : float
        phase shift of grating in degrees
    intensities : Sequence[float, float] or None (default)
        min and max intensity of sine-wave
    target_indices : int, or Sequence[int, ...]
        indices segments where targets will be placed
    intensity_target : float, or Sequence[float, ...], optional
        intensity value for each target, by default 0.5.
        Can specify as many intensities as number of target_indices;
        If fewer intensities are passed than target_indices, cycles through intensities
    origin : "corner", "mean" or "center"
        if "corner": set origin to upper left corner (default)
        if "mean": set origin to hypothetical image center
        if "center": set origin to real center (closest existing value to mean)
    round_phase_width : Bool
        if True, round width of bars given resolution

    Returns
    -------
    dict[str, Any]
        dict with the stimulus (key: "img"),
        mask with integer index for each target (key: "target_mask"),
        and additional keys containing stimulus parameters
    """
    if len(intensities) != 2:
        raise ValueError("intensities should be [float, float]")

    lst = [visual_size, ppd, shape, frequency, n_bars, bar_width]
    if len([x for x in lst if x is not None]) < 3:
        raise ValueError(
            "'grating' needs 3 non-None arguments for resolving from 'visual_size', "
            "'ppd', 'shape', 'frequency', 'n_bars', 'bar_width'"
        )

    if rotation % 180 == 0.0:
        distance_metric = "horizontal"
    elif rotation % 180 == 90.0:
        distance_metric = "vertical"
    else:
        distance_metric = "oblique"

    # Spatial square-wave grating
    stim = waves.sine(
        visual_size=visual_size,
        ppd=ppd,
        shape=shape,
        frequency=frequency,
        n_phases=n_bars,
        phase_width=bar_width,
        period=period,
        rotation=rotation,
        phase_shift=phase_shift,
        intensities=intensities,
        origin=origin,
        round_phase_width=round_phase_width,
        distance_metric=distance_metric,
    )

    # Repackage output
    stim["n_bars"] = stim.pop("n_phases")
    stim["bar_width"] = stim.pop("phase_width")
    stim.pop("distance_metric")

    # Rename mask
    stim["bar_mask"] = stim["grating_mask"]

    # Add targets
    stim = place_targets(
        stim=stim,
        element_mask_key="grating_mask",
        target_indices=target_indices,
        intensity_target=intensity_target,
    )
    return stim


def square_linear(
    visual_size=None,
    ppd=None,
    shape=None,
    frequency=None,
    n_bars=None,
    bar_width=None,
    period="ignore",
    rotation=0.0,
    phase_shift=0,
    intensity_bars=(0.0, 1.0),
    target_indices=(),
    intensity_target=0.5,
    origin="corner",
    round_phase_width=True,
):
    """Linear (horizontal, vertical, oblique) square-wave grating (set of bars), with some bar(s) as target(s)

    Parameters
    ----------
    visual_size : Sequence[Number, Number], Number, or None (default)
        visual size [height, width] of image, in degrees
    ppd : Sequence[Number, Number], Number, or None (default)
        pixels per degree [vertical, horizontal]
    shape : Sequence[Number, Number], Number, or None (default)
        shape [height, width] of image, in pixels
    frequency : Number, or None (default)
        spatial frequency of grating, in cycles per degree visual angle
    n_bars : int, or None (default)
        number of bars in the grating
    bar_width : Number, or None (default)
        width of a single bar, in degrees visual angle
    period : "even", "odd", "either" or "ignore" (default)
        ensure whether the grating has "even" number of phases, "odd"
        number of phases, either or whether not to round the number of
        phases ("ignore")
    rotation : float, optional
        rotation (in degrees), counterclockwise, by default 0.0 (horizontal)
    phase_shift : float
        phase shift of grating in degrees
    intensity_bars : Sequence[float, ...]
        intensity value for each bar, by default (1.0, 0.0).
        Can specify as many intensities as n_bars;
        If fewer intensities are passed than n_bars, cycles through intensities
    target_indices : int, or Sequence[int, ...]
        indices segments where targets will be placed
    intensity_target : float, or Sequence[float, ...], optional
        intensity value for each target, by default 0.5.
        Can specify as many intensities as number of target_indices;
        If fewer intensities are passed than target_indices, cycles through intensities
    origin : "corner", "mean" or "center"
        if "corner": set origin to upper left corner (default)
        if "mean": set origin to hypothetical image center
        if "center": set origin to real center (closest existing value to mean)
    round_phase_width : Bool
        if True, round width of bars given resolution

    Returns
    -------
    dict[str, Any]
        dict with the stimulus (key: "img"),
        mask with integer index for each target (key: "target_mask"),
        and additional keys containing stimulus parameters
    """
    lst = [visual_size, ppd, shape, frequency, n_bars, bar_width]
    if len([x for x in lst if x is not None]) < 3:
        raise ValueError(
            "'grating' needs 3 non-None arguments for resolving from 'visual_size', "
            "'ppd', 'shape', 'frequency', 'n_bars', 'bar_width'"
        )

    # Spatial square-wave grating
    stim = waves.square(
        visual_size=visual_size,
        ppd=ppd,
        shape=shape,
        frequency=frequency,
        n_phases=n_bars,
        phase_width=bar_width,
        period=period,
        rotation=rotation,
        phase_shift=phase_shift,
        origin=origin,
        round_phase_width=round_phase_width,
        distance_metric="oblique",
        intensities=intensity_bars,
    )

    # Repackage output
    stim["n_bars"] = stim.pop("n_phases")
    stim["bar_width"] = stim.pop("phase_width")
    stim["intensity_bars"] = stim.pop("intensities")
    stim.pop("distance_metric")

    # Rename mask
    stim["bar_mask"] = stim["grating_mask"]

    # Add targets
    stim = place_targets(
        stim=stim,
        element_mask_key="grating_mask",
        target_indices=target_indices,
        intensity_target=intensity_target,
    )
    return stim


def staircase_linear(
    visual_size=None,
    ppd=None,
    shape=None,
    frequency=None,
    n_bars=None,
    bar_width=None,
    period="ignore",
    rotation=0.0,
    phase_shift=0,
    intensity_bars=(0.0, 1.0),
    target_indices=(),
    intensity_target=0.5,
    origin="corner",
    round_phase_width=True,
):
    """Linear staircase, with some bar(s) as target(s)

    Parameters
    ----------
    visual_size : Sequence[Number, Number], Number, or None (default)
        visual size [height, width] of image, in degrees
    ppd : Sequence[Number, Number], Number, or None (default)
        pixels per degree [vertical, horizontal]
    shape : Sequence[Number, Number], Number, or None (default)
        shape [height, width] of image, in pixels
    frequency : Number, or None (default)
        spatial frequency of grating, in cycles per degree visual angle
    n_bars : int, or None (default)
        number of bars in the grating
    bar_width : Number, or None (default)
        width of a single bar, in degrees visual angle
    period : "even", "odd", "either" or "ignore" (default)
        ensure whether the grating has "even" number of phases, "odd"
        number of phases, either or whether not to round the number of
        phases ("ignore")
    rotation : float, optional
        rotation (in degrees), counterclockwise, by default 0.0 (horizontal)
    phase_shift : float
        phase shift of grating in degrees
    intensity_bars : Sequence[float, float] or Sequence[float, ...]
        min and max intensity of sine-wave, by default (0.0, 1.0).
        Can also specify as many intensities as n_bars
    target_indices : int, or Sequence[int, ...]
        indices segments where targets will be placed
    intensity_target : float, or Sequence[float, ...], optional
        intensity value for each target, by default 0.5.
        Can specify as many intensities as number of target_indices;
        If fewer intensities are passed than target_indices, cycles through intensities
    origin : "corner", "mean" or "center"
        if "corner": set origin to upper left corner (default)
        if "mean": set origin to hypothetical image center
        if "center": set origin to real center (closest existing value to mean)
    round_phase_width : Bool
        if True, round width of bars given resolution
    intensity_bars : Sequence[float, ...]
        if len(intensity_bars)==2, intensity range of staircase (default 0.0, 1.0);
        if len(intensity_bars)>2, intensity value for each bar.
        Can specify as many intensity_bars as n_bars.
        If fewer intensity_bars are passed than n_bars, cycles through intensities.


    Returns
    -------
    dict[str, Any]
        dict with the stimulus (key: "img"),
        mask with integer index for each target (key: "target_mask"),
        and additional keys containing stimulus parameters
    """
    lst = [visual_size, ppd, shape, frequency, n_bars, bar_width]
    if len([x for x in lst if x is not None]) < 3:
        raise ValueError(
            "'grating' needs 3 non-None arguments for resolving from 'visual_size', "
            "'ppd', 'shape', 'frequency', 'n_bars', 'bar_width'"
        )

    # Spatial square-wave grating
    stim = waves.staircase(
        visual_size=visual_size,
        ppd=ppd,
        shape=shape,
        frequency=frequency,
        n_phases=n_bars,
        phase_width=bar_width,
        period=period,
        rotation=rotation,
        phase_shift=phase_shift,
        origin=origin,
        round_phase_width=round_phase_width,
        distance_metric="oblique",
        intensities=intensity_bars,
    )

    # Repackage output
    stim["n_bars"] = stim.pop("n_phases")
    stim["bar_width"] = stim.pop("phase_width")
    stim["intensity_bars"] = stim.pop("intensities")
    stim.pop("distance_metric")

    # Rename mask
    stim["bar_mask"] = stim["grating_mask"]

    # Add targets
    stim = place_targets(
        stim=stim,
        element_mask_key="grating_mask",
        target_indices=target_indices,
        intensity_target=intensity_target,
    )
    return stim


def sine_radial(
    visual_size=None,
    ppd=None,
    shape=None,
    frequency=None,
    n_rings=None,
    ring_width=None,
    period="ignore",
    rotation=0.0,
    phase_shift=0,
    intensities=(0.0, 1.0),
    target_indices=(),
    intensity_target=0.5,
    origin="mean",
    round_phase_width=True,
    clip=False,
    intensity_background=0.5,
):
    """Circular sine-wave grating (set of rings) over the whole image, with some ring(s) as target(s)

    Parameters
    ----------
    visual_size : Sequence[Number, Number], Number, or None (default)
        visual size [height, width] of image, in degrees
    ppd : Sequence[Number, Number], Number, or None (default)
        pixels per degree [vertical, horizontal]
    shape : Sequence[Number, Number], Number, or None (default)
        shape [height, width] of image, in pixels
    frequency : Number, or None (default)
        spatial frequency of grating, in cycles per degree visual angle
    n_rings : int, or None (default)
        number of rings
    ring_width : Number, or None (default)
        width of a single ring, in degrees
    period : "even", "odd", "either" or "ignore" (default)
        ensure whether the grating has "even" number of phases, "odd"
        number of phases, either or whether not to round the number of
        phases ("ignore")
    rotation : float, optional
        rotation (in degrees), counterclockwise, by default 0.0 (horizontal)
    phase_shift : float
        phase shift of grating in degrees
    intensities : Sequence[float, float] or None (default)
        min and max intensity of sine-wave
    target_indices : int, or Sequence[int, ...]
        indices segments where targets will be placed
    intensity_target : float, or Sequence[float, ...], optional
        intensity value for each target, by default 0.5.
        Can specify as many intensities as number of target_indices;
        If fewer intensities are passed than target_indices, cycles through intensities
    origin : "corner", "mean" or "center"
        if "corner": set origin to upper left corner
        if "mean": set origin to hypothetical image center (default)
        if "center": set origin to real center (closest existing value to mean)
    round_phase_width : Bool
        if True, round width of rings given resolution
    clip : Bool
        if True, clip stimulus to image size (default: False)
    intensity_background : float (optional)
        intensity value of background (if clipped), by default 0.5

    Returns
    -------
    dict[str, Any]
        dict with the stimulus (key: "img"),
        mask with integer index for each target (key: "target_mask"),
        and additional keys containing stimulus parameters
    """
    if len(intensities) != 2:
        raise ValueError("intensity_rings should be [float, float]")

    lst = [visual_size, ppd, shape, frequency, n_rings, ring_width]
    if len([x for x in lst if x is not None]) < 3:
        raise ValueError(
            "'grating' needs 3 non-None arguments for resolving from 'visual_size', "
            "'ppd', 'shape', 'frequency', 'n_rings', 'rings_width'"
        )

    # Spatial square-wave grating
    stim = waves.sine(
        visual_size=visual_size,
        ppd=ppd,
        shape=shape,
        frequency=frequency,
        n_phases=n_rings,
        phase_width=ring_width,
        period=period,
        rotation=rotation,
        phase_shift=phase_shift,
        intensities=intensities,
        origin=origin,
        round_phase_width=round_phase_width,
        distance_metric="radial",
    )

    # Repackage output
    stim["n_rings"] = stim.pop("n_phases")
    stim["ring_width"] = stim.pop("phase_width")
    stim.pop("distance_metric")

    # Clip?
    if clip:
        csize = min(stim["visual_size"]) / 2.0
        circle = disc(
            visual_size=stim["visual_size"],
            ppd=stim["ppd"],
            radius=csize,
            origin=origin,
        )
        stim["img"] = np.where(circle["ring_mask"], stim["img"], intensity_background)
        stim["grating_mask"] = np.where(circle["ring_mask"], stim["grating_mask"], 0)
    stim["clip"] = clip
    stim["intensity_background"] = intensity_background

    # Rename mask
    stim["ring_mask"] = stim["grating_mask"]

    # Add targets
    stim = place_targets(
        stim=stim,
        element_mask_key="grating_mask",
        target_indices=target_indices,
        intensity_target=intensity_target,
    )
    return stim


def square_radial(
    visual_size=None,
    ppd=None,
    shape=None,
    frequency=None,
    n_rings=None,
    ring_width=None,
    period="ignore",
    rotation=0.0,
    phase_shift=0,
    intensity_rings=(0.0, 1.0),
    target_indices=(),
    intensity_target=0.5,
    origin="mean",
    round_phase_width=True,
    clip=False,
    intensity_background=0.5,
):
    """Circular square-wave grating (set of rings) over the whole image, with some ring(s) as target(s)

    Parameters
    ----------
    visual_size : Sequence[Number, Number], Number, or None (default)
        visual size [height, width] of image, in degrees
    ppd : Sequence[Number, Number], Number, or None (default)
        pixels per degree [vertical, horizontal]
    shape : Sequence[Number, Number], Number, or None (default)
        shape [height, width] of image, in pixels
    frequency : Number, or None (default)
        spatial frequency of grating, in cycles per degree visual angle
    n_rings : int, or None (default)
        number of rings in the grating
    ring_width : Number, or None (default)
        width of a single ring, in degrees visual angle
    period : "even", "odd", "either" or "ignore" (default)
        ensure whether the grating has "even" number of phases, "odd"
        number of phases, either or whether not to round the number of
        phases ("ignore")
    rotation : float, optional
        rotation (in degrees), counterclockwise, by default 0.0 (horizontal)
    phase_shift : float
        phase shift of grating in degrees
    intensity_rings : Sequence[float, float]
        intensity value for each ring, by default (1.0, 0.0).
    target_indices : int, or Sequence[int, ...]
        indices segments where targets will be placed
    intensity_target : float, or Sequence[float, ...], optional
        intensity value for each target, by default 0.5.
        Can specify as many intensities as number of target_indices;
        If fewer intensities are passed than target_indices, cycles through intensities
    origin : "corner", "mean" or "center"
        if "corner": set origin to upper left corner
        if "mean": set origin to hypothetical image center (default)
        if "center": set origin to real center (closest existing value to mean)
    round_phase_width : Bool
        if True, round width of rings given resolution
    clip : Bool
        if True, clip stimulus to image size (default: False)
    intensity_background : float (optional)
        intensity value of background (if clipped), by default 0.5

    Returns
    -------
    dict[str, Any]
        dict with the stimulus (key: "img"),
        mask with integer index for each target (key: "target_mask"),
        and additional keys containing stimulus parameters
    """
    lst = [visual_size, ppd, shape, frequency, n_rings, ring_width]
    if len([x for x in lst if x is not None]) < 3:
        raise ValueError(
            "'grating' needs 3 non-None arguments for resolving from 'visual_size', "
            "'ppd', 'shape', 'frequency', 'n_rings', 'ring_width'"
        )

    # Spatial square-wave grating
    stim = waves.square(
        visual_size=visual_size,
        ppd=ppd,
        shape=shape,
        frequency=frequency,
        n_phases=n_rings,
        phase_width=ring_width,
        period=period,
        rotation=rotation,
        phase_shift=phase_shift,
        origin=origin,
        round_phase_width=round_phase_width,
        distance_metric="radial",
        intensities=intensity_rings,
    )

    # Repackage output
    stim["n_rings"] = stim.pop("n_phases")
    stim["ring_width"] = stim.pop("phase_width")
    stim["intensity_rings"] = stim.pop("intensities")
    stim.pop("distance_metric")

    # Clip?
    if clip:
        csize = min(stim["visual_size"]) / 2.0
        circle = disc(
            visual_size=stim["visual_size"],
            ppd=stim["ppd"],
            radius=csize,
            origin=origin,
        )
        stim["img"] = np.where(circle["ring_mask"], stim["img"], intensity_background)
        stim["grating_mask"] = np.where(circle["ring_mask"], stim["grating_mask"], 0)
    stim["clip"] = clip
    stim["intensity_background"] = intensity_background

    # Rename mask
    stim["ring_mask"] = stim["grating_mask"]

    # Add targets
    stim = place_targets(
        stim=stim,
        element_mask_key="grating_mask",
        target_indices=target_indices,
        intensity_target=intensity_target,
    )
    return stim


def staircase_radial(
    visual_size=None,
    ppd=None,
    shape=None,
    frequency=None,
    n_rings=None,
    ring_width=None,
    period="ignore",
    rotation=0.0,
    phase_shift=0,
    intensity_rings=(0.0, 1.0),
    target_indices=(),
    intensity_target=0.5,
    origin="mean",
    round_phase_width=True,
    clip=False,
    intensity_background=0.5,
):
    """Radial staircase, with some ring(s) as target(s)

    Parameters
    ----------
    visual_size : Sequence[Number, Number], Number, or None (default)
        visual size [height, width] of image, in degrees
    ppd : Sequence[Number, Number], Number, or None (default)
        pixels per degree [vertical, horizontal]
    shape : Sequence[Number, Number], Number, or None (default)
        shape [height, width] of image, in pixels
    frequency : Number, or None (default)
        spatial frequency of grating, in cycles per degree visual angle
    n_rings : int, or None (default)
        number of rings in the grating
    ring_width : Number, or None (default)
        width of a single ring, in degrees visual angle
    period : "even", "odd", "either" or "ignore" (default)
        ensure whether the grating has "even" number of phases, "odd"
        number of phases, either or whether not to round the number of
        phases ("ignore")
    rotation : float, optional
        rotation (in degrees), counterclockwise, by default 0.0 (horizontal)
    phase_shift : float
        phase shift of grating in degrees
    intensity_rings : Sequence[float, ...]
        if len(intensity_rings)==2, intensity range of staircase (default 0.0, 1.0);
        if len(intensity_rings)>2, intensity value for each ring.
        Can specify as many intensity_rings as n_rings.
        If fewer intensity_bars are passed than n_rings, cycles through intensities.
    target_indices : int, or Sequence[int, ...]
        indices segments where targets will be placed
    intensity_target : float, or Sequence[float, ...], optional
        intensity value for each target, by default 0.5.
        Can specify as many intensities as number of target_indices;
        If fewer intensities are passed than target_indices, cycles through intensities
    origin : "corner", "mean" or "center"
        if "corner": set origin to upper left corner
        if "mean": set origin to hypothetical image center (default)
        if "center": set origin to real center (closest existing value to mean)
    round_phase_width : Bool
        if True, round width of rings given resolution
    clip : Bool
        if True, clip stimulus to image size (default: False)
    intensity_background : float (optional)
        intensity value of background (if clipped), by default 0.5

    Returns
    -------
    dict[str, Any]
        dict with the stimulus (key: "img"),
        mask with integer index for each target (key: "target_mask"),
        and additional keys containing stimulus parameters
    """
    lst = [visual_size, ppd, shape, frequency, n_rings, ring_width]
    if len([x for x in lst if x is not None]) < 3:
        raise ValueError(
            "'grating' needs 3 non-None arguments for resolving from 'visual_size', "
            "'ppd', 'shape', 'frequency', 'n_rings', 'ring_width'"
        )

    # Spatial square-wave grating
    stim = waves.staircase(
        visual_size=visual_size,
        ppd=ppd,
        shape=shape,
        frequency=frequency,
        n_phases=n_rings,
        phase_width=ring_width,
        period=period,
        rotation=rotation,
        phase_shift=phase_shift,
        origin=origin,
        round_phase_width=round_phase_width,
        distance_metric="radial",
        intensities=intensity_rings,
    )

    # Repackage output
    stim["n_rings"] = stim.pop("n_phases")
    stim["ring_width"] = stim.pop("phase_width")
    stim["intensity_rings"] = stim.pop("intensities")
    stim.pop("distance_metric")

    # Clip?
    if clip:
        csize = min(stim["visual_size"]) / 2.0
        circle = disc(
            visual_size=stim["visual_size"],
            ppd=stim["ppd"],
            radius=csize,
            origin=origin,
        )
        stim["img"] = np.where(circle["ring_mask"], stim["img"], intensity_background)
        stim["grating_mask"] = np.where(circle["ring_mask"], stim["grating_mask"], 0)
    stim["clip"] = clip
    stim["intensity_background"] = intensity_background

    # Rename mask
    stim["ring_mask"] = stim["grating_mask"]

    # Add targets
    stim = place_targets(
        stim=stim,
        element_mask_key="grating_mask",
        target_indices=target_indices,
        intensity_target=intensity_target,
    )
    return stim


def sine_rectilinear(
    visual_size=None,
    ppd=None,
    shape=None,
    frequency=None,
    n_frames=None,
    frame_width=None,
    period="ignore",
    rotation=0.0,
    phase_shift=0,
    intensities=(0.0, 1.0),
    target_indices=(),
    intensity_target=0.5,
    origin="mean",
    round_phase_width=True,
    clip=False,
    intensity_background=0.5,
):
    """Rectilinear sine-wave grating (set of frames) over the whole image, with some frame(s) as target(s)

    Parameters
    ----------
    visual_size : Sequence[Number, Number], Number, or None (default)
        visual size [height, width] of image, in degrees
    ppd : Sequence[Number, Number], Number, or None (default)
        pixels per degree [vertical, horizontal]
    shape : Sequence[Number, Number], Number, or None (default)
        shape [height, width] of image, in pixels
    frequency : Number, or None (default)
        spatial frequency of grating, in cycles per degree visual angle
    n_frames : int, or None (default)
        number of frames
    frame_width : Number, or None (default)
        width of a single frame, in degrees
    period : "even", "odd", "either" or "ignore" (default)
        ensure whether the grating has "even" number of phases, "odd"
        number of phases, either or whether not to round the number of
        phases ("ignore")
    rotation : float, optional
        rotation (in degrees), counterclockwise, by default 0.0 (horizontal)
    phase_shift : float
        phase shift of grating in degrees
    intensities : Sequence[float, float] or None (default)
        min and max intensity of sine-wave
    target_indices : int, or Sequence[int, ...]
        indices segments where targets will be placed
    intensity_target : float, or Sequence[float, ...], optional
        intensity value for each target, by default 0.5.
        Can specify as many intensities as number of target_indices;
        If fewer intensities are passed than target_indices, cycles through intensities
    origin : "corner", "mean" or "center"
        if "corner": set origin to upper left corner
        if "mean": set origin to hypothetical image center (default)
        if "center": set origin to real center (closest existing value to mean)
    round_phase_width : Bool
        if True, round width of frames given resolution
    clip : Bool
        if True, clip stimulus to image size (default: False)
    intensity_background : float (optional)
        intensity value of background (if clipped), by default 0.5

    Returns
    -------
    dict[str, Any]
        dict with the stimulus (key: "img"),
        mask with integer index for each target (key: "target_mask"),
        and additional keys containing stimulus parameters
    """
    if len(intensities) != 2:
        raise ValueError("intensity_frames should be [float, float]")

    lst = [visual_size, ppd, shape, frequency, n_frames, frame_width]
    if len([x for x in lst if x is not None]) < 3:
        raise ValueError(
            "'grating' needs 3 non-None arguments for resolving from 'visual_size', "
            "'ppd', 'shape', 'frequency', 'n_frames', 'frames_width'"
        )

    # Spatial square-wave grating
    stim = waves.sine(
        visual_size=visual_size,
        ppd=ppd,
        shape=shape,
        frequency=frequency,
        n_phases=n_frames,
        phase_width=frame_width,
        period=period,
        rotation=rotation,
        phase_shift=phase_shift,
        intensities=intensities,
        origin=origin,
        round_phase_width=round_phase_width,
        distance_metric="rectilinear",
    )

    # Repackage output
    stim["n_frames"] = stim.pop("n_phases")
    stim["frame_width"] = stim.pop("phase_width")
    stim.pop("distance_metric")

    # Clip?
    if clip:
        if origin == "corner":
            rsize = min(stim["visual_size"]) / 2
            rect = rectangle(
                visual_size=stim["visual_size"],
                ppd=stim["ppd"],
                rectangle_size=rsize,
                rectangle_position=(0, 0),
            )
        else:
            rsize = min(stim["visual_size"])
            rect = rectangle(
                visual_size=stim["visual_size"],
                ppd=stim["ppd"],
                rectangle_size=rsize,
            )
        stim["img"] = np.where(rect["rectangle_mask"], stim["img"], intensity_background)
        stim["grating_mask"] = np.where(rect["rectangle_mask"], stim["grating_mask"], 0)
    stim["clip"] = clip
    stim["intensity_background"] = intensity_background

    # Rename mask
    stim["frame_mask"] = stim["grating_mask"]

    # Add targets
    stim = place_targets(
        stim=stim,
        element_mask_key="grating_mask",
        target_indices=target_indices,
        intensity_target=intensity_target,
    )
    return stim


def square_rectilinear(
    visual_size=None,
    ppd=None,
    shape=None,
    frequency=None,
    n_frames=None,
    frame_width=None,
    period="ignore",
    rotation=0.0,
    phase_shift=0,
    intensity_frames=(0.0, 1.0),
    target_indices=(),
    intensity_target=0.5,
    origin="mean",
    round_phase_width=True,
    clip=False,
    intensity_background=0.5,
):
    """Rectilinear square-wave grating (set of frames) over the whole image, with
    some frame(s) as target(s)

    Parameters
    ----------
    visual_size : Sequence[Number, Number], Number, or None (default)
        visual size [height, width] of image, in degrees
    ppd : Sequence[Number, Number], Number, or None (default)
        pixels per degree [vertical, horizontal]
    shape : Sequence[Number, Number], Number, or None (default)
        shape [height, width] of image, in pixels
    frequency : Number, or None (default)
        spatial frequency of grating, in cycles per degree visual angle
    n_frames : int, or None (default)
        number of frames in the grating
    frame_width : Number, or None (default)
        width of a single frame, in degrees visual angle
    period : "even", "odd", "either" or "ignore" (default)
        ensure whether the grating has "even" number of phases, "odd"
        number of phases, either or whether not to round the number of
        phases ("ignore")
    rotation : float, optional
        rotation (in degrees), counterclockwise, by default 0.0 (horizontal)
    phase_shift : float
        phase shift of grating in degrees
    intensity_frames : Sequence[float, float]
        intensity value for each frame, by default (1.0, 0.0).
    target_indices : int, or Sequence[int, ...]
        indices segments where targets will be placed
    intensity_target : float, or Sequence[float, ...], optional
        intensity value for each target, by default 0.5.
        Can specify as many intensities as number of target_indices;
        If fewer intensities are passed than target_indices, cycles through intensities
    origin : "corner", "mean" or "center"
        if "corner": set origin to upper left corner
        if "mean": set origin to hypothetical image center (default)
        if "center": set origin to real center (closest existing value to mean)
    round_phase_width : Bool
        if True, round width of frames given resolution
    clip : Bool
        if True, clip stimulus to image size (default: False)
    intensity_background : float (optional)
        intensity value of background (if clipped), by default 0.5

    Returns
    -------
    dict[str, Any]
        dict with the stimulus (key: "img"),
        mask with integer index for each target (key: "target_mask"),
        and additional keys containing stimulus parameters
    """
    lst = [visual_size, ppd, shape, frequency, n_frames, frame_width]
    if len([x for x in lst if x is not None]) < 3:
        raise ValueError(
            "'grating' needs 3 non-None arguments for resolving from 'visual_size', "
            "'ppd', 'shape', 'frequency', 'n_frames', 'frame_width'"
        )

    # Spatial square-wave grating
    stim = waves.square(
        visual_size=visual_size,
        ppd=ppd,
        shape=shape,
        frequency=frequency,
        n_phases=n_frames,
        phase_width=frame_width,
        period=period,
        rotation=rotation,
        phase_shift=phase_shift,
        origin=origin,
        round_phase_width=round_phase_width,
        distance_metric="rectilinear",
        intensities=intensity_frames,
    )

    # Repackage output
    stim["n_frames"] = stim.pop("n_phases")
    stim["frame_width"] = stim.pop("phase_width")
    stim["intensity_frames"] = stim.pop("intensities")
    stim.pop("distance_metric")

    # Clip?
    if clip:
        if origin == "corner":
            rsize = min(stim["visual_size"]) / 2
            rect = rectangle(
                visual_size=stim["visual_size"],
                ppd=stim["ppd"],
                rectangle_size=rsize,
                rectangle_position=(0, 0),
            )
        else:
            rsize = min(stim["visual_size"])
            rect = rectangle(
                visual_size=stim["visual_size"],
                ppd=stim["ppd"],
                rectangle_size=rsize,
            )
        stim["img"] = np.where(rect["rectangle_mask"], stim["img"], intensity_background)
        stim["grating_mask"] = np.where(rect["rectangle_mask"], stim["grating_mask"], 0)
    stim["clip"] = clip
    stim["intensity_background"] = intensity_background

    # Rename mask
    stim["frame_mask"] = stim["grating_mask"]

    # Add targets
    stim = place_targets(
        stim=stim,
        element_mask_key="grating_mask",
        target_indices=target_indices,
        intensity_target=intensity_target,
    )
    return stim


def staircase_rectilinear(
    visual_size=None,
    ppd=None,
    shape=None,
    frequency=None,
    n_frames=None,
    frame_width=None,
    period="ignore",
    rotation=0.0,
    phase_shift=0,
    intensity_frames=(0.0, 1.0),
    target_indices=(),
    intensity_target=0.5,
    origin="mean",
    round_phase_width=True,
    clip=False,
    intensity_background=0.5,
):
    """Rectilinear staircase, with some frame(s) as target(s)

    Parameters
    ----------
    visual_size : Sequence[Number, Number], Number, or None (default)
        visual size [height, width] of image, in degrees
    ppd : Sequence[Number, Number], Number, or None (default)
        pixels per degree [vertical, horizontal]
    shape : Sequence[Number, Number], Number, or None (default)
        shape [height, width] of image, in pixels
    frequency : Number, or None (default)
        spatial frequency of grating, in cycles per degree visual angle
    n_frames : int, or None (default)
        number of frames in the grating
    frame_width : Number, or None (default)
        width of a single frame, in degrees visual angle
    period : "even", "odd", "either" or "ignore" (default)
        ensure whether the grating has "even" number of phases, "odd"
        number of phases, either or whether not to round the number of
        phases ("ignore")
    rotation : float, optional
        rotation (in degrees), counterclockwise, by default 0.0 (horizontal)
    phase_shift : float
        phase shift of grating in degrees
    intensity_frames : Sequence[float, ...]
        if len(intensity_frames)==2, intensity range of staircase (default 0.0, 1.0);
        if len(intensity_frames)>2, intensity value for each frame.
        Can specify as many intensity_frames as n_frames.
        If fewer intensity_frames are passed than n_frames, cycles through intensities.
    target_indices : int, or Sequence[int, ...]
        indices segments where targets will be placed
    intensity_target : float, or Sequence[float, ...], optional
        intensity value for each target, by default 0.5.
        Can specify as many intensities as number of target_indices;
        If fewer intensities are passed than target_indices, cycles through intensities
    origin : "corner", "mean" or "center"
        if "corner": set origin to upper left corner
        if "mean": set origin to hypothetical image center (default)
        if "center": set origin to real center (closest existing value to mean)
    round_phase_width : Bool
        if True, round width of frames given resolution
    clip : Bool
        if True, clip stimulus to image size (default: False)
    intensity_background : float (optional)
        intensity value of background (if clipped), by default 0.5

    Returns
    -------
    dict[str, Any]
        dict with the stimulus (key: "img"),
        mask with integer index for each target (key: "target_mask"),
        and additional keys containing stimulus parameters
    """
    lst = [visual_size, ppd, shape, frequency, n_frames, frame_width]
    if len([x for x in lst if x is not None]) < 3:
        raise ValueError(
            "'grating' needs 3 non-None arguments for resolving from 'visual_size', "
            "'ppd', 'shape', 'frequency', 'n_frames', 'frame_width'"
        )

    # Spatial square-wave grating
    stim = waves.staircase(
        visual_size=visual_size,
        ppd=ppd,
        shape=shape,
        frequency=frequency,
        n_phases=n_frames,
        phase_width=frame_width,
        period=period,
        rotation=rotation,
        phase_shift=phase_shift,
        origin=origin,
        round_phase_width=round_phase_width,
        distance_metric="rectilinear",
        intensities=intensity_frames,
    )

    # Repackage output
    stim["n_frames"] = stim.pop("n_phases")
    stim["frame_width"] = stim.pop("phase_width")
    stim["intensity_frames"] = stim.pop("intensities")
    stim.pop("distance_metric")

    # Clip?
    if clip:
        if origin == "corner":
            rsize = min(stim["visual_size"]) / 2
            rect = rectangle(
                visual_size=stim["visual_size"],
                ppd=stim["ppd"],
                rectangle_size=rsize,
                rectangle_position=(0, 0),
            )
        else:
            rsize = min(stim["visual_size"])
            rect = rectangle(
                visual_size=stim["visual_size"],
                ppd=stim["ppd"],
                rectangle_size=rsize,
            )
        stim["img"] = np.where(rect["rectangle_mask"], stim["img"], intensity_background)
        stim["grating_mask"] = np.where(rect["rectangle_mask"], stim["grating_mask"], 0)
    stim["clip"] = clip
    stim["intensity_background"] = intensity_background

    # Rename mask
    stim["frame_mask"] = stim["grating_mask"]

    # Add targets
    stim = place_targets(
        stim=stim,
        element_mask_key="grating_mask",
        target_indices=target_indices,
        intensity_target=intensity_target,
    )
    return stim


def sine_angular(
    visual_size=None,
    ppd=None,
    shape=None,
    frequency=None,
    n_segments=None,
    segment_width=None,
    period="ignore",
    rotation=0.0,
    phase_shift=0,
    intensities=(0.0, 1.0),
    target_indices=(),
    intensity_target=0.5,
    origin="mean",
    round_phase_width=True,
):
    """Angular sine-wave grating (set of segments) over the whole image, with some segment(s) as target(s)

    Parameters
    ----------
    visual_size : Sequence[Number, Number], Number, or None (default)
        visual size [height, width] of image, in degrees
    ppd : Sequence[Number, Number], Number, or None (default)
        pixels per degree [vertical, horizontal]
    shape : Sequence[Number, Number], Number, or None (default)
        shape [height, width] of image, in pixels
    frequency : Number, or None (default)
        spatial frequency of grating, in cycles per image
    n_segments : int, or None (default)
        number of segments
    segment_width : Number, or None (default)
        width of a single segment, in degrees
    period : "even", "odd", "either" or "ignore" (default)
        ensure whether the grating has "even" number of phases, "odd"
        number of phases, either or whether not to round the number of
        phases ("ignore")
    rotation : float, optional
        rotation (in degrees), counterclockwise, by default 0.0 (horizontal)
    phase_shift : float
        phase shift of grating in degrees
    intensities : Sequence[float, float] or None (default)
        min and max intensity of sine-wave
    target_indices : int, or Sequence[int, ...]
        indices segments where targets will be placed
    intensity_target : float, or Sequence[float, ...], optional
        intensity value for each target, by default 0.5.
        Can specify as many intensities as number of target_indices;
        If fewer intensities are passed than target_indices, cycles through intensities
    origin : "corner", "mean" or "center"
        if "corner": set origin to upper left corner (default)
        if "mean": set origin to hypothetical image center
        if "center": set origin to real center (closest existing value to mean)
    round_phase_width : Bool
        if True, round width of segments given resolution

    Returns
    -------
    dict[str, Any]
        dict with the stimulus (key: "img"),
        mask with integer index for each target (key: "target_mask"),
        and additional keys containing stimulus parameters
    """
    if len(intensities) != 2:
        raise ValueError("intensity_segments should be [float, float]")

    lst = [visual_size, ppd, shape, frequency, n_segments, segment_width]
    if len([x for x in lst if x is not None]) < 3:
        raise ValueError(
            "'grating' needs 3 non-None arguments for resolving from 'visual_size', "
            "'ppd', 'shape', 'frequency', 'n_segments', 'segments_width'"
        )

    # Spatial square-wave grating
    stim = waves.sine(
        visual_size=visual_size,
        ppd=ppd,
        shape=shape,
        frequency=frequency,
        n_phases=n_segments,
        phase_width=segment_width,
        period=period,
        rotation=rotation,
        phase_shift=phase_shift,
        intensities=intensities,
        origin=origin,
        round_phase_width=round_phase_width,
        distance_metric="angular",
    )

    # Repackage output
    stim["n_segments"] = stim.pop("n_phases")
    stim["segment_width"] = stim.pop("phase_width")
    stim.pop("distance_metric")

    # Rename mask
    stim["segment_mask"] = stim["grating_mask"]

    # Add targets
    stim = place_targets(
        stim=stim,
        element_mask_key="grating_mask",
        target_indices=target_indices,
        intensity_target=intensity_target,
    )
    return stim


def square_angular(
    visual_size=None,
    ppd=None,
    shape=None,
    frequency=None,
    n_segments=None,
    segment_width=None,
    period="ignore",
    rotation=0.0,
    phase_shift=0,
    intensity_segments=(0.0, 1.0),
    target_indices=(),
    intensity_target=0.5,
    origin="mean",
    round_phase_width=True,
):
    """Angular square-wave grating (set of segments) over the whole image, with some segment(s) as target(s)

    Parameters
    ----------
    visual_size : Sequence[Number, Number], Number, or None (default)
        visual size [height, width] of image, in degrees
    ppd : Sequence[Number, Number], Number, or None (default)
        pixels per degree [vertical, horizontal]
    shape : Sequence[Number, Number], Number, or None (default)
        shape [height, width] of image, in pixels
    frequency : Number, or None (default)
        spatial frequency of grating, in cycles per image
    n_segments : int, or None (default)
        number of segments in the grating
    segment_width : Number, or None (default)
        width of a single segment, in degrees visual angle
    period : "even", "odd", "either" or "ignore" (default)
        ensure whether the grating has "even" number of phases, "odd"
        number of phases, either or whether not to round the number of
        phases ("ignore")
    rotation : float, optional
        rotation (in degrees), counterclockwise, by default 0.0 (horizontal)
    phase_shift : float
        phase shift of grating in degrees
    intensity_segments : Sequence[float, ...]
        intensity value for each segment, by default (1.0, 0.0).
        Can specify as many intensities as n_segments;
        If fewer intensities are passed than n_segments, cycles through intensities
    target_indices : int, or Sequence[int, ...]
        indices segments where targets will be placed
    intensity_target : float, or Sequence[float, ...], optional
        intensity value for each target, by default 0.5.
        Can specify as many intensities as number of target_indices;
        If fewer intensities are passed than target_indices, cycles through intensities
    origin : "corner", "mean" or "center"
        if "corner": set origin to upper left corner (default)
        if "mean": set origin to hypothetical image center
        if "center": set origin to real center (closest existing value to mean)
    round_phase_width : Bool
        if True, round width of segments given resolution

    Returns
    -------
    dict[str, Any]
        dict with the stimulus (key: "img"),
        mask with integer index for each target (key: "target_mask"),
        and additional keys containing stimulus parameters
    """
    lst = [visual_size, ppd, shape, frequency, n_segments, segment_width]
    if len([x for x in lst if x is not None]) < 3:
        raise ValueError(
            "'grating' needs 3 non-None arguments for resolving from 'visual_size', "
            "'ppd', 'shape', 'frequency', 'n_segments', 'segment_width'"
        )

    # Spatial square-wave grating
    stim = waves.square(
        visual_size=visual_size,
        ppd=ppd,
        shape=shape,
        frequency=frequency,
        n_phases=n_segments,
        phase_width=segment_width,
        period=period,
        rotation=rotation,
        phase_shift=phase_shift,
        origin=origin,
        round_phase_width=round_phase_width,
        distance_metric="angular",
        intensities=intensity_segments,
    )

    # Repackage output
    stim["n_segments"] = stim.pop("n_phases")
    stim["segment_width"] = stim.pop("phase_width")
    stim["intensity_segments"] = stim.pop("intensities")
    stim.pop("distance_metric")

    # Rename mask
    stim["segment_mask"] = stim["grating_mask"]

    # Add targets
    stim = place_targets(
        stim=stim,
        element_mask_key="grating_mask",
        target_indices=target_indices,
        intensity_target=intensity_target,
    )
    return stim


def staircase_angular(
    visual_size=None,
    ppd=None,
    shape=None,
    frequency=None,
    n_segments=None,
    segment_width=None,
    period="ignore",
    rotation=0.0,
    phase_shift=0,
    intensity_segments=(0.0, 1.0),
    target_indices=(),
    intensity_target=0.5,
    origin="center",
    round_phase_width=True,
):
    """Angular staircase, with some segment(s) as target(s)

    Parameters
    ----------
    visual_size : Sequence[Number, Number], Number, or None (default)
        visual size [height, width] of image, in degrees
    ppd : Sequence[Number, Number], Number, or None (default)
        pixels per degree [vertical, horizontal]
    shape : Sequence[Number, Number], Number, or None (default)
        shape [height, width] of image, in pixels
    frequency : Number, or None (default)
        spatial frequency of grating, in cycles per image
    n_segments : int, or None (default)
        number of segments in the grating
    segment_width : Number, or None (default)
        width of a single segment, in degrees visual angle
    period : "even", "odd", "either" or "ignore" (default)
        ensure whether the grating has "even" number of phases, "odd"
        number of phases, either or whether not to round the number of
        phases ("ignore")
    rotation : float, optional
        rotation (in degrees), counterclockwise, by default 0.0 (horizontal)
    phase_shift : float
        phase shift of grating in degrees
    intensity_segments : Sequence[float, ...]
        if len(intensity_segments)==2, intensity range of staircase (default 0.0, 1.0);
        if len(intensity_segments)>2, intensity value for each segment.
        Can specify as many intensity_segments as n_segments.
        If fewer intensity_segments are passed than n_segments, cycles through intensities.
    target_indices : int, or Sequence[int, ...]
        indices segments where targets will be placed
    intensity_target : float, or Sequence[float, ...], optional
        intensity value for each target, by default 0.5.
        Can specify as many intensities as number of target_indices;
        If fewer intensities are passed than target_indices, cycles through intensities
    origin : "corner", "mean" or "center"
        if "corner": set origin to upper left corner (default)
        if "mean": set origin to hypothetical image center
        if "center": set origin to real center (closest existing value to mean)
    round_phase_width : Bool
        if True, round width of segments given resolution

    Returns
    -------
    dict[str, Any]
        dict with the stimulus (key: "img"),
        mask with integer index for each target (key: "target_mask"),
        and additional keys containing stimulus parameters
    """
    lst = [visual_size, ppd, shape, frequency, n_segments, segment_width]
    if len([x for x in lst if x is not None]) < 3:
        raise ValueError(
            "'grating' needs 3 non-None arguments for resolving from 'visual_size', "
            "'ppd', 'shape', 'frequency', 'n_segments', 'segment_width'"
        )

    # Spatial square-wave grating
    stim = waves.staircase(
        visual_size=visual_size,
        ppd=ppd,
        shape=shape,
        frequency=frequency,
        n_phases=n_segments,
        phase_width=segment_width,
        period=period,
        rotation=rotation,
        phase_shift=phase_shift,
        origin=origin,
        round_phase_width=round_phase_width,
        distance_metric="angular",
        intensities=intensity_segments,
    )

    # Repackage output
    stim["n_segments"] = stim.pop("n_phases")
    stim["segment_width"] = stim.pop("phase_width")
    stim["intensity_segments"] = stim.pop("intensities")
    stim.pop("distance_metric")

    # Rename mask
    stim["segment_mask"] = stim["grating_mask"]

    # Add targets
    stim = place_targets(
        stim=stim,
        element_mask_key="grating_mask",
        target_indices=target_indices,
        intensity_target=intensity_target,
    )
    return stim


def overview(**kwargs):
    """Generate example stimuli from this module

    Returns
    -------
    stims : dict
        dict with all stimuli containing individual stimulus dicts.
    """
    default_params = {"visual_size": 15, "ppd": 30}
    default_params.update(kwargs)

    grating_params = {
        "period": "ignore",
        "phase_shift": 0,
        "round_phase_width": False,
    }

    grating_params2 = {
        "period": "ignore",
        "phase_shift": 0,
        "round_phase_width": False,
        "target_indices": (2, 5),
    }

    # fmt: off
    stimuli = {
        "sine_wave_linear": sine_linear(**default_params, **grating_params, bar_width=1, rotation=45),
        "sine_wave_radial": sine_radial(**default_params, **grating_params, ring_width=1),
        "sine_wave_rectilinear": sine_rectilinear(**default_params, **grating_params, frame_width=1),
        "sine_wave_angular": sine_angular(**default_params, **grating_params, n_segments=10),


        "square_wave_linear": square_linear(**default_params, **grating_params, bar_width=1, rotation=45),
        "square_wave_radial": square_radial(**default_params, **grating_params, ring_width=1),
        "square_wave_rectilinear": square_rectilinear(**default_params, **grating_params, frame_width=1, clip=True),
        "square_wave_angular": square_angular(**default_params, **grating_params, n_segments=10),

        "staircase_linear": staircase_linear(**default_params, **grating_params, frequency=0.4),
        "staircase_radial": staircase_radial(**default_params, **grating_params, frequency=0.4),
        "staircase_rectilinear": staircase_rectilinear(**default_params, **grating_params, frequency=0.4),
        "staircase_angular": staircase_angular(**default_params, **grating_params, n_segments=10),

        "sine_wave_linear_with_targets": sine_linear(**default_params, **grating_params2, bar_width=1, rotation=45),
        "sine_wave_radial_with_targets": sine_radial(**default_params, **grating_params2, ring_width=1),
        "sine_wave_rectilinear_with_targets": sine_rectilinear(**default_params, **grating_params2, frame_width=1),
        "sine_wave_angular_with_targets": sine_angular(**default_params, **grating_params2, n_segments=10),


        "square_wave_linear_with_targets": square_linear(**default_params, **grating_params2, bar_width=1, rotation=45),
        "square_wave_radial_with_targets": square_radial(**default_params, **grating_params2, ring_width=1),
        "square_wave_rectilinear_with_targets": square_rectilinear(**default_params, **grating_params2, frame_width=1, clip=True),
        "square_wave_angular_with_targets": square_angular(**default_params, **grating_params2, n_segments=10),

        "staircase_linear_with_targets": staircase_linear(**default_params, **grating_params2, frequency=0.4),
        "staircase_radial_with_targets": staircase_radial(**default_params, **grating_params2, frequency=0.4),
        "staircase_rectilinear_with_targets": staircase_rectilinear(**default_params, **grating_params2, frequency=0.4),
        "staircase_angular_with_targets": staircase_angular(**default_params, **grating_params2, n_segments=10),
    }
    # fmt: on

    return stimuli


if __name__ == "__main__":
    from stimupy.utils import plot_stimuli

    stims = overview()
    plot_stimuli(stims, mask=False, save=None)



"""Example paper module (citation information here)

This set only serves as an example / template
for how to set up a stimulus-set-module

Attributes
----------
__all__ (list of str): list of all stimulus-functions,
    these are exported by this module when executing
        >>> from stimupy.papers.example import *

References
----------
reference information here

"""

from stimupy.components import combine_masks, draw_regions, shapes

# Define original size resolution parameters
VISUAL_SIZE = (10, 12)
PPD = 10

__all__ = [
    "my_bullseye",
    "my_inverse_bullseye",
]


# Helper function:
def bullseye_geometry(ppd=PPD):
    """Helper function to create the bullseye geometry

    This function itself is not a stimulus,
    and will not be shown in `stimupy.papers.my_paper`

    Parameters
    ----------
    ppd : int
        Resolution of stimulus in pixels per degree. (default: 10)

    """
    # Create center (target) disc:
    disc = shapes.disc(
        visual_size=VISUAL_SIZE, ppd=ppd, radius=2, intensity_disc=0.5, intensity_background=0.5
    )

    # Create first ring, white:
    ring_1 = shapes.ring(
        visual_size=VISUAL_SIZE, ppd=ppd, radii=(2, 3), intensity_ring=1, intensity_background=0.5
    )

    # Create second ring, black:
    ring_2 = shapes.ring(
        visual_size=VISUAL_SIZE, ppd=ppd, radii=(3, 4), intensity_ring=0, intensity_background=0.5
    )

    bullseye_mask = combine_masks(disc["ring_mask"], ring_1["ring_mask"], ring_2["ring_mask"])

    return bullseye_mask


# New stimulus function:
def my_bullseye(ppd=PPD):
    """My bullseye stimulus: grey on white on black

    Parameters
    ----------
    ppd : int
        Resolution of stimulus in pixels per degree. (default: 32)

    Returns
    -------
    dict of str
        dict with the stimulus (key: "img")
        and additional keys containing stimulus parameters

    """

    # Call geometry helper function
    bullseye_mask = bullseye_geometry(ppd=ppd)

    bullseye_img = draw_regions(
        mask=bullseye_mask, intensities=[0.5, 1, 0], intensity_background=0.5
    )

    # Package into stim-dict, adding parameter information
    stim = {
        "img": bullseye_img,
        "visual_size": VISUAL_SIZE,
        "ppd": ppd,
        "radii": (2, 3, 4),
        "intensities": (0.5, 1, 0),
        "intensity_background": 0.5,
    }

    # Output
    return stim


# Second stimulus function:
def my_inverse_bullseye(ppd=PPD):
    """My other bullseye stimulus: grey on black on white


    Parameters
    ----------
    ppd : int
        Resolution of stimulus in pixels per degree. (default: 10)

    Returns
    -------
    dict of str
        dict with the stimulus (key: "img")
        and additional keys containing stimulus parameters

    """

    # Call geometry helper function
    bullseye_mask = bullseye_geometry(ppd=ppd)

    bullseye_img = draw_regions(
        mask=bullseye_mask, intensities=[0.5, 0, 1], intensity_background=0.5
    )

    # Package into stim-dict, adding parameter information
    stim = {
        "img": bullseye_img,
        "visual_size": VISUAL_SIZE,
        "ppd": ppd,
        "radii": (2, 3, 4),
        "intensities": (0.5, 0, 1),
        "intensity_background": 0.5,
        "note": "Here is some additional information: I like this stimulus!",
    }

    # Output
    return stim



import numpy as np

from stimupy.utils import resolution

__all__ = [
    "varying_cells",
    "cube",
]


def varying_cells(
    ppd=None,
    cell_lengths=None,
    cell_thickness=None,
    cell_spacing=None,
    target_indices=(),
    intensity_background=0.0,
    intensity_cells=1.0,
    intensity_target=0.5,
):
    """
    Cube stimulus (Agostini & Galmonte, 2002) with flexible cell lengths.

    Parameters
    ----------
    ppd : Number or None (default)
        pixels per degree (visual angle)
    cell_lengths : Sequence[Number, ...], Number of None (default)
        lengths of individual cells in degrees
    cell_thickness : Number or None (default)
        thickness of each cell in degrees
    cell_spacing : Number or None (default)
        spacing between cells in degrees
    target_indices : Sequence or None
        target indices; will be used on each side
    intensity_background : float
        intensity value for background
    intensity_cells : float
        intensity value for grid cells
    intensity_target : float
        intensity value for target

    Returns
    -------
    dict[str, Any]
        dict with the stimulus (key: "img"),
        mask with integer index for each target (key: "target_mask"),
        and additional keys containing stimulus parameters

    References
    ----------
    Agostini, T., and Galmonte, A. (2002).
        Perceptual organization overcomes the effects of local surround
        in determining simultaneous lightness contrast.
        Psychol. Sci. 13, 89-93.
        https://doi.org/10.1111/1467-9280.00417
    Domijan, D. (2015).
        A neurocomputational account
        of the role of contour facilitation in brightness perception.
        Frontiers in Human Neuroscience, 9, 93.
        https://doi.org/10.3389/fnhum.2015.00093
    """
    if not isinstance(ppd, (float, int)):
        ppd = np.unique(ppd)
        if len(ppd) != 1:
            raise ValueError("ppd has to be the same in x and y direction")
        else:
            ppd = ppd[0]

    if isinstance(cell_lengths, (float, int)):
        cell_lengths = (cell_lengths,)
    if target_indices is None:
        target_indices = ()
    if isinstance(target_indices, (float, int)):
        target_indices = (target_indices,)

    n_cells = len(cell_lengths)

    clengths = resolution.lengths_from_visual_angles_ppd(cell_lengths, ppd)
    cthick = resolution.lengths_from_visual_angles_ppd(cell_thickness, ppd)
    cspace = resolution.lengths_from_visual_angles_ppd(cell_spacing, ppd)
    height = (
        np.maximum(clengths[0], cthick)
        + np.maximum(clengths[-1], cthick)
        + sum(clengths[1 : n_cells - 1])
        + cspace * (n_cells - 1)
    )
    width = height

    # Initiate image
    cell_mask = np.zeros([height, width])
    target_mask = np.zeros([height, width])

    xs = 0
    counter = 1
    for i in range(n_cells):
        if i in target_indices:
            fill_mask = 1
        else:
            fill_mask = 0

        # Add cells top
        cell_mask[0:cthick, xs : xs + clengths[i]] = counter
        target_mask[0:cthick, xs : xs + clengths[i]] += fill_mask

        # Add cells bottom
        cell_mask[height - cthick : :, width - xs - clengths[i] : width - xs] = counter + 1
        target_mask[height - cthick : :, width - xs - clengths[i] : width - xs] += fill_mask

        # Add cells left
        cell_mask[height - xs - clengths[i] : height - xs, 0:cthick] = counter + 2
        target_mask[height - xs - clengths[i] : height - xs, 0:cthick] += fill_mask

        # Add cells right
        cell_mask[xs : xs + clengths[i], width - cthick : :] = counter + 3
        target_mask[xs : xs + clengths[i], width - cthick : :] += fill_mask

        if i == 0:
            xs += np.maximum(clengths[i], cthick) + cspace
        elif i == n_cells - 2:
            xs += np.maximum(clengths[-2], cthick) + cspace
        else:
            xs += clengths[i] + cspace
        counter += 4

    unique_vals = np.unique(cell_mask)
    for v in range(len(unique_vals) - 1):
        cell_mask[cell_mask == unique_vals[v + 1]] = v + 1

    img = np.where(cell_mask != 0, intensity_cells, intensity_background)
    img = np.where(target_mask != 0, intensity_target, img)
    target_mask = np.where(target_mask >= 1, 1, 0)

    stim = {
        "img": img,
        "cell_mask": cell_mask.astype(int),
        "target_mask": target_mask.astype(int),
        "shape": img.shape,
        "visual_size": np.array(img.shape) / ppd,
        "ppd": ppd,
        "target_indices": target_indices,
        "cell_lengths": cell_lengths,
        "cell_thickness": cell_thickness,
        "cell_spacing": cell_spacing,
        "intensity_background": intensity_background,
        "intensity_cells": intensity_cells,
        "intensity_target": intensity_target,
    }
    return stim


def cube(
    visual_size=None,
    ppd=None,
    shape=None,
    n_cells=None,
    target_indices=(),
    cell_thickness=None,
    cell_spacing=None,
    intensity_background=0.0,
    intensity_cells=1.0,
    intensity_target=0.5,
):
    """Cube illusion (Agostini & Galmonte, 2002)

    Parameters
    ----------
    visual_size : Sequence[Number, Number], Number, or None (default)
        visual size [height, width] of image, in degrees
    ppd : Sequence[Number, Number], Number, or None (default)
        pixels per degree [vertical, horizontal]
    shape : Sequence[Number, Number], Number, or None (default)
        shape [height, width] of image, in pixels
    n_cells : int
        the number of square cells (not counting background) per dimension
    target_indices : Sequence
        Target indices. Will be used on each side
    cell_thickness : Number or None (default)
        thickness of each cell in degrees
    cell_spacing : Sequence[Number, Number], Number or None (default)
        spacing between cells in degrees (height, width)
    intensity_background : float
        intensity value for background
    intensity_cells : float
        intensity value for grid cells
    intensity_target : float
        intensity value for target

    Returns
    -------
    dict[str, Any]
        dict with the stimulus (key: "img"),
        mask with integer index for each target (key: "target_mask"),
        and additional keys containing stimulus parameters

    References
    ----------
    Agostini, T., and Galmonte, A. (2002).
        Perceptual organization overcomes the effects of local surround
        in determining simultaneous lightness contrast.
        Psychol. Sci. 13, 89-93.
        https://doi.org/10.1111/1467-9280.00417
    Domijan, D. (2015).
        A neurocomputational account
        of the role of contour facilitation in brightness perception.
        Frontiers in Human Neuroscience, 9, 93.
        https://doi.org/10.3389/fnhum.2015.00093
    """

    # Resolve resolution
    shape, visual_size, ppd = resolution.resolve(shape=shape, visual_size=visual_size, ppd=ppd)
    if len(np.unique(ppd)) > 1:
        raise ValueError("ppd should be equal in x- and y-direction")
    if n_cells is None:
        raise ValueError("cube() missing argument 'n_cells' which is not 'None'")
    if cell_thickness is None:
        raise ValueError("cube() missing argument 'cell_thickness' which is not 'None'")
    if cell_spacing is None:
        raise ValueError("cube() missing argument 'cell_spacing' which is not 'None'")

    if isinstance(cell_spacing, (float, int)):
        cell_spacing = (cell_spacing, cell_spacing)
    if isinstance(n_cells, (float, int)):
        n_cells = (n_cells, n_cells)
    if target_indices is None:
        target_indices = ()
    if isinstance(target_indices, (float, int)):
        target_indices = (target_indices,)

    height, width = shape
    cell_space = resolution.lengths_from_visual_angles_ppd(cell_spacing, np.unique(ppd))
    cell_thick = resolution.lengths_from_visual_angles_ppd(cell_thickness, np.unique(ppd))

    # Initiate image
    cell_mask = np.zeros([height, width])
    target_mask = np.zeros([height, width])

    # Calculate cell widths and heights
    cell_height = int((height - cell_space[0] * (n_cells[0] - 1)) / n_cells[0])
    cell_width = int((width - cell_space[1] * (n_cells[1] - 1)) / n_cells[1])

    if (cell_thick > cell_height) or (cell_thick > cell_width):
        raise ValueError("cannot fit all cells into image")

    # Calculate cell placements:
    xs = np.arange(0, width - 1, cell_width + cell_space[1])
    rxs = xs[::-1]
    ys = np.arange(0, height - 1, cell_height + cell_space[0])
    rys = ys[::-1]

    # Add cells: top and bottom
    counter = 1
    for i in range(n_cells[1]):
        if i in target_indices:
            fill_mask = 1
        else:
            fill_mask = 0

        cell_mask[0:cell_thick, xs[i] : xs[i] + cell_width] = counter
        cell_mask[height - cell_thick : :, rxs[i] : rxs[i] + cell_width] = counter + 1
        target_mask[0:cell_thick, xs[i] : xs[i] + cell_width] += fill_mask
        target_mask[height - cell_thick : :, rxs[i] : rxs[i] + cell_width] += fill_mask
        counter += 2

    # Add cells: left and right
    for i in range(n_cells[0]):
        if i in target_indices:
            fill_mask = 1
        else:
            fill_mask = 0

        cell_mask[rys[i] : rys[i] + cell_height, 0:cell_thick] = counter
        cell_mask[ys[i] : ys[i] + cell_height, height - cell_thick : :] = counter + 1
        target_mask[rys[i] : rys[i] + cell_height, 0:cell_thick] += fill_mask
        target_mask[ys[i] : ys[i] + cell_height, height - cell_thick : :] += fill_mask
        counter += 2

    unique_vals = np.unique(cell_mask)
    for v in range(len(unique_vals) - 1):
        cell_mask[cell_mask == unique_vals[v + 1]] = v + 1

    img = np.where(cell_mask != 0, intensity_cells, intensity_background)
    img = np.where(target_mask != 0, intensity_target, img)
    target_mask = np.where(target_mask >= 1, 1, 0)

    stim = {
        "img": img,
        "cell_mask": cell_mask.astype(int),
        "target_mask": target_mask.astype(int),
        "shape": shape,
        "visual_size": visual_size,
        "ppd": ppd,
        "target_indices": target_indices,
        "n_cells": n_cells,
        "cell_thickness": cell_thickness,
        "cell_spacing": cell_spacing,
        "intensity_background": intensity_background,
        "intensity_cells": intensity_cells,
        "intensity_target": intensity_target,
    }
    return stim


def overview(**kwargs):
    """Generate example stimuli from this module

    Returns
    -------
    stims : dict
        dict with all stimuli containing individual stimulus dicts.
    """
    default_params = {
        "ppd": 30,
        "cell_thickness": 1,
        "cell_spacing": 0.5,
    }
    default_params.update(kwargs)

    # fmt: off
    stimuli = {
        "cubes_regular": cube(**default_params, visual_size=10, n_cells=5, target_indices=(1, 2)),
        "cubes_variable": varying_cells(**default_params, cell_lengths=(2, 4, 2), target_indices=1),
    }
    # fmt: on

    return stimuli


if __name__ == "__main__":
    from stimupy.utils import plot_stimuli

    stims = overview()
    plot_stimuli(stims, mask=False, save=None)



import copy
import json
import pickle
from hashlib import md5
from pathlib import Path

import numpy as np
from PIL import Image
from scipy.io import savemat

__all__ = [
    "array_to_checksum",
    "array_to_image",
    "array_to_npy",
    "array_to_mat",
    "array_to_pickle",
    "arrays_to_checksum",
    "to_json",
    "to_mat",
    "to_pickle",
]


def array_to_checksum(arr):
    """Hash (md5) array values, and return hex-checksum

    Parameters
    ----------
    arr : numpy.ndarray
        array to be hashed.

    Returns
    ----------
    str
        hex-string representation of hash (MD5) of given array
    """
    return md5(np.ascontiguousarray(arr.round(8))).hexdigest()


def array_to_image(arr, filename, format=None, norm=True):
    """Save a 2D numpy array as a grayscale image file.

    Parameters
    ----------
    arr : numpy.ndarray
        array to be exported. Values will be cropped to [0,255].
    filename : Path or str
        (full) path to the file to be created.
    norm : bool
        multiply array by 255, by default True
    """
    try:
        arr_to_write = np.ascontiguousarray(arr)
    except TypeError as e:
        raise e from ValueError("arr should be a numpy.ndarray(-like)")

    if norm:
        arr_to_write = arr_to_write * 255

    im = Image.fromarray(arr_to_write.astype("uint8"), mode="L")
    im.save(filename)


def array_to_npy(arr, filename):
    """Save a numpy array to npy-file.

    Parameters
    ----------
    arr : numpy.ndarray
        array to be exported.
    filename : Path or str
        (full) path to the file to be created.
    """
    try:
        arr_to_write = np.ascontiguousarray(arr)
    except TypeError as e:
        raise e from ValueError("arr should be a numpy.ndarray(-like)")

    filepath = Path(filename).resolve().with_suffix(".npy")

    np.save(filepath, arr_to_write)


def array_to_mat(arr, filename):
    """Save a numpy array to a mat-file.

    Parameters
    ----------
    arr : numpy.ndarray
        array to be exported.
    filename : Path or str
        (full) path to the file to be created.
    """
    try:
        arr_to_write = np.ascontiguousarray(arr)
    except TypeError as e:
        raise e from ValueError("arr should be a numpy.ndarray(-like)")

    filepath = Path(filename).resolve().with_suffix(".mat")

    savemat(filepath, {"arr": arr_to_write})


def array_to_pickle(arr, filename):
    """Save a numpy array to a pickle-file.

    Parameters
    ----------
    arr : numpy.ndarray
        array to be exported.
    filename : Path or str
        (full) path to the file to be created.
    """
    try:
        arr_to_write = np.ascontiguousarray(arr)
    except TypeError as e:
        raise e from ValueError("arr should be a numpy.ndarray(-like)")

    filepath = Path(filename).resolve().with_suffix(".pickle")

    with filepath.open("wb") as file:
        pickle.dump({"arr": arr_to_write}, file, protocol=pickle.HIGHEST_PROTOCOL)


def array_to_json(arr, filename):
    """Save a numpy array to a (pretty) JSON.

    Parameters
    ----------
    arr : numpy.ndarray
        array to be exported.
    filename : Path or str
        (full) path to the file to be created.
    """
    try:
        arr_to_write = np.ascontiguousarray(arr)
    except TypeError as e:
        raise e from ValueError("arr should be a numpy.ndarray(-like)")

    filepath = Path(filename).resolve().with_suffix(".json")

    with filepath.open("w", encoding="utf-8") as file:
        json.dump(arr_to_write.tolist(), file, ensure_ascii=False, indent=4)


def arrays_to_checksum(stim, keys=["img", "mask"]):
    """Hash (md5) values of arrays specified in keys, and save only the hex

    Parameters
    ----------
    stim : dict
        stimulus dictionary to export.
    keys : str of list of str
        keys of dict for which the hashing should be performed

    Returns
    ----------
    dict[str, Any]
        same as input dict but keys now only contain the hex
    """
    if isinstance(keys, str):
        keys = [
            keys,
        ]

    for key in keys:
        stim[key] = array_to_checksum(stim[key])

    return stim


def to_json(stim, filename):
    """Save stimulus-dict(s) as (pretty) JSON

    Parameters
    ----------
    stim : dict
        stimulus dictionary to export.
    filename : Path or str
        (full) path to the file to be created.

    """
    if not isinstance(stim, dict):
        raise ValueError("stim should be a dict")

    filepath = Path(filename).resolve().with_suffix(".json")

    stim2 = copy.deepcopy(stim)
    for key in stim2.keys():
        # np.ndarrays are not serializable; change to list
        if isinstance(stim2[key], np.ndarray):
            stim2[key] = stim2[key].tolist()

    with filepath.open("w", encoding="utf-8") as file:
        json.dump(stim2, file, ensure_ascii=False, indent=4)


def to_mat(stim, filename):
    """Save stimulus-dict(s) as mat-file

    Parameters
    ----------
    stim : dict
        stimulus dictionary to export.
    filename : Path or str
        (full) path to the file to be created.

    """
    if not isinstance(stim, dict):
        raise ValueError("stim should be a dict")

    filepath = Path(filename).resolve().with_suffix(".mat")

    savemat(filepath, stim)


def to_pickle(stim, filename):
    """Save stimulus-dict(s) as pickle-file

    Parameters
    ----------
    stim : dict
        stimulus dictionary to export.
    filename : Path or str
        (full) path to the file to be created.

    """
    if not isinstance(stim, dict):
        raise ValueError("stim should be a dict")

    filepath = Path(filename).resolve().with_suffix(".pickle")

    stim2 = copy.deepcopy(stim)
    for key in stim2.keys():
        # certain classes can cause problems for pickles; change to list
        if key in ["visual_size", "ppd", "shape"]:
            stim2[key] = list(stim2[key])

    with filepath.open("wb") as file:
        pickle.dump(stim2, file, protocol=pickle.HIGHEST_PROTOCOL)



import numpy as np

__all__ = [
    "avg_target_values",
    "avg_img_values",
    "all_img_values",
    "img_values",
]


def avg_target_values(stim, mask_key="target_mask", f_average=np.median):
    """Average pixel value in each target region of stimulus

    Parameters
    ----------
    stim : dict[str: Any]
        stimulus-dict with at least "img" and "mask"
        containing the stimulus image and integer-mask, respectively.
    mask_key : str
        string with mask-key name
    f_average: function, default=numpy.median
        How to average/summarise the pixels in each target region

    Returns
    -------
    list[float]
        each entry in the list is the average value of pixels in target region,
        index in the list is the integer index in the mask

    See Also
    --------
    avg_img_values
    """
    return avg_img_values(image=stim["img"], mask=stim[mask_key], f_average=f_average)


def avg_img_values(image, mask, f_average=np.median):
    """Average values of pixels in image, per target region in integer mask

    Values are calculated as means or medians, depending on the mode.

    Parameters
    ----------
    image : 2D numpy array
        2D numpy array containing pixel values of the image
    mask : 2D numpy array
        2D numpy array of same size as image. Each target patch has an integer value.
        Each pixel inside this patch has this integer value.
        Patches do not need to be continuous.
    f_average: function, default=numpy.median
        How to average/summarise the pixels in each target region

    Returns
    -------
    list[float]
        each entry in the list is the average value of pixels in target region,
        index in the list is the integer index in the mask

    See Also
    --------
    all_img_values
    """
    masked_outputs = all_img_values(image, mask)
    values = [f_average(o[np.isfinite(o)]) for o in masked_outputs]

    return values


def all_img_values(img, mask):
    """Isolate all image values/pixels, per target region specified in integer mask

    Parameters
    ----------
    img : numpy.ndarray
        Image-array of pixel values to be masked
    mask : numpy.ndarray
        Array of same size as img.
        Each region-of-interest in mask is represented by an integer index.
        Each pixel inside this patch has this integer value.
        Patches do not need to be contiguous.

    Returns
    -------
    list[numpy.arrays]
        Each image/element of the list is a numpy.ndarray representing an image.
        There is one image for each target patch in the integer mask.
        In each image all values are set to NaN
        except the ones corresponding to the target values of the respective target patch.

    See Also
    --------
    img_values

    """
    idc = np.unique(mask.astype(int))
    imgs = [img_values(img, mask == idx) for idx in idc]
    return imgs


def img_values(img, mask):
    """Isolate only image pixels specified by a binary mask

    Parameters
    ----------
    img : numpy.ndarray
        Image-array of pixel values to be masked
    mask : numpy.ndarray
        Array of same size as img.
        All non-zero pixels/values are treated as ones in a binary bit mask.

    Returns
    -------
    numpy.ndarray
        numpy.ndarray of same size as img.
        All bits corresponding to zero bits in the mask are set to NaN.
    """

    return np.where(mask, img, np.nan)



"""Stimuli from Robinson, Hammon & de Sa (2007)
https://doi.org/10.1016/j.visres.2007.02.017

This module reproduces all of the stimuli used by Robinson,
Hammon & de Sa (2007) as they were provided to the model described
in that paper.

Each stimulus is provided by a separate function,
a full list can be found as stimupy.papers.RHS2007.__all__

The output of each of these functions is a stimulus dictionary.

For a visual representation of all the stimuli and their mask,
simply run this module as a script:

    $ python stimuli/papers/RHS2007.py

Attributes
----------
__all__ (list of str): list of all stimulus-functions
    that are exported by this module when executing
        >>> from stimupy.papers.RHS2007 import *

References
----------
Robinson, A. E., Hammon, P. S., & de Sa, V. R. (2007).
    Explaining brightness illusions
    using spatial filtering and local response normalization.
    Vision Research, 47(12), 1631-1644.
    https://doi.org/10.1016/j.visres.2007.02.017
"""

import numpy as np

import stimupy
from stimupy.utils import (
    flip_dict,
    pad_dict_by_visual_size,
    pad_dict_to_shape,
    pad_dict_to_visual_size,
    resolution,
    rotate_dict,
    stack_dicts,
)

__all__ = [
    "WE_thick",
    "WE_thin_wide",
    "WE_dual",
    "WE_anderson",
    "WE_howe",
    "WE_zigzag",
    "WE_radial_thick_small",
    "WE_radial_thick",
    "WE_radial_thin_small",
    "WE_radial_thin",
    "WE_circular1",
    "WE_circular05",
    "WE_circular025",
    "grating_induction",
    "sbc_large",
    "sbc_small",
    "todorovic_equal",
    "todorovic_in_large",
    "todorovic_in_small",
    "todorovic_out",
    "checkerboard_016",
    "checkerboard_094",
    "checkerboard_21",
    "corrugated_mondrian",
    "benary_cross",
    "todorovic_benary1_2",
    "todorovic_benary3_4",
    "todorovic_benary1_2_3_4",
    "bullseye_thin",
    "bullseye_thick",
]

VISEXTENT = (32.0, 32.0)
PPD = 32
v1, v2, v3 = 0.0, 0.5, 1.0


def gen_all(ppd=PPD, pad=True, skip=False):
    stims = {}  # save the stimulus-dicts in a larger dict, with name as key
    for stim_name in __all__:
        print(f"Generating RHS2007.{stim_name}")

        # Get a reference to the actual function
        func = globals()[stim_name]
        try:
            stim = func(ppd=ppd, pad=pad)

            # Accumulate
            stims[stim_name] = stim
        except NotImplementedError as e:
            if not skip:
                raise e
            # Skip stimuli that aren't implemented
            print("-- not implemented")
            pass

    return stims


def WE_thick(ppd=PPD, pad=True):
    """White stimulus as shown in Robinson, Hammon, & de Sa (2007) Fig 1a.
    Grating size: 12x16 deg
    Grating frequency: 0.25 cpd
    Target size: 4x2 deg

    Parameters
    ----------
    ppd : int
        Resolution of stimulus in pixels per degree. (default: 32)
    pad : bool
        If True, include padding to 32x32 deg (default: True)

    Returns
    -------
    dict of str
        dict with the stimulus (key: "img") and target mask (key: "target_mask")
        and additional keys containing stimulus parameters

    References
    ----------
    Blakeslee, B., & McCourt, M. E. (1999).
        A multiscale spatial ﬁltering account
        of the White eﬀect, simultaneous brightness contrast and grating induction.
        Vision Research, 39, 4361-4377.
    White, M. (1979).
        A new eﬀect of pattern on perceived lightness.
        Perception, 8, 413-416.
    """

    height, width = 12.0, 16.0
    params = {
        "ppd": ppd,
        "frequency": 4.0 / width,
        "intensity_bars": (v3, v1),
        "intensity_target": v2,
        "target_indices": (3, 6),
        "target_heights": 4.0,
        "period": "even",
    }

    stim = stimupy.whites.white(
        visual_size=(height, width),
        **params,
    )

    if pad:
        stim = pad_dict_to_visual_size(stim, VISEXTENT, ppd, pad_value=v2)

    experimental_data = {
        "effect_strength": 4.18,
    }

    params.update(
        visual_size=np.array(stim["img"].shape) / ppd,
        shape=stim["img"].shape,
        intensity_range=(v1, v3),
        experimental_data=experimental_data,
    )
    return {**stim, **params}


def WE_thin_wide(ppd=PPD, pad=True):
    """White stimulus as shown in Robinson, Hammon, & de Sa (2007) Fig 1b.
    Grating size: 12x16 deg
    Grating frequency: 0.5 cpd
    Target size: 2x1 deg

    Parameters
    ----------
    ppd : int
        Resolution of stimulus in pixels per degree. (default: 32)
    pad : bool
        If True, include padding to 32x32 deg (default: True)

    Returns
    -------
    dict of str
        dict with the stimulus (key: "img") and target mask (key: "target_mask")
        and additional keys containing stimulus parameters

    References
    ----------
    Blakeslee, B., & McCourt, M. E. (1999).
        A multiscale spatial ﬁltering account
        of the White eﬀect, simultaneous brightness contrast and grating induction.
        Vision Research, 39, 4361-4377.
    White, M. (1979).
        A new eﬀect of pattern on perceived lightness. Perception, 8, 413-416.
    """

    height, width = 12.0, 16.0
    params = {
        "ppd": ppd,
        "frequency": 8.0 / width,
        "intensity_bars": (v1, v3),
        "intensity_target": v2,
        "target_indices": (4, 13),
        "target_heights": 2.0,
        "period": "even",
    }

    stim = stimupy.whites.white(
        visual_size=(height, width),
        **params,
    )

    if pad:
        stim = pad_dict_to_visual_size(stim, VISEXTENT, ppd, pad_value=v2)

    experimental_data = {
        "effect_strength": 4.6,
    }

    params.update(
        visual_size=np.array(stim["img"].shape) / ppd,
        shape=stim["img"].shape,
        intensity_range=(v1, v3),
        experimental_data=experimental_data,
    )
    return {**stim, **params}


def WE_dual(ppd=PPD, pad=True):
    """Dual White stimulus as shown in Robinson, Hammon, & de Sa (2007) Fig 1c.
    Grating sizes: 6x8 deg
    Grating frequency: 0.5 cpd
    Target size: 2x1 deg

    Parameters
    ----------
    ppd : int
        Resolution of stimulus in pixels per degree. (default: 32)
    pad : bool
        If True, include padding to 32x32 deg (default: True)

    Returns
    -------
    dict of str
        dict with the stimulus (key: "img") and target mask (key: "target_mask")
        and additional keys containing stimulus parameters
    """

    height, width = 6.0, 8.0
    params = {
        "ppd": ppd,
        "frequency": 4.0 / width,
        "intensity_bars": (v3, v1),
        "intensity_target": v2,
        "target_indices": (3, 6),
        "target_heights": 2.0,
        "period": "even",
    }

    stim1 = stimupy.whites.white(
        visual_size=(height, width),
        **params,
    )

    stim2 = stimupy.whites.white(
        visual_size=(height, width),
        **params,
    )
    stim2 = rotate_dict(stim2, 3)

    if pad:
        shape = np.array(resolution.shape_from_visual_size_ppd(VISEXTENT, ppd)) / (1, 2)
    else:
        shape = np.max([stim1["img"].shape, stim2["img"].shape], axis=1)

    # Pad and stack
    stim1 = pad_dict_to_shape(stim1, shape, pad_value=v2)
    stim2 = pad_dict_to_shape(stim2, shape, pad_value=v2)
    stim = stack_dicts(stim1, stim2, keep_mask_indices=False)

    params.update(
        visual_size=np.array(stim["img"].shape) / ppd,
        shape=stim["img"].shape,
        intensity_range=(v1, v3),
    )
    return {**stim, **params}


def WE_anderson(ppd=PPD, pad=True):
    """Anderson variation of White stimulus as shown in Robinson, Hammon, & de Sa (2007) Fig 1d.
    Grating size: 16x16 deg
    Grating frequency: 0.5 cpd
    Target size: 3.2x1 deg

    Parameters
    ----------
    ppd : int
        Resolution of stimulus in pixels per degree. (default: 32)
    pad : bool
        If True, include padding to 32x32 deg (default: True)

    Returns
    -------
    dict of str
        dict with the stimulus (key: "img") and target mask (key: "target_mask")
        and additional keys containing stimulus parameters

    References
    ----------
    Anderson, B. L. (2001).
        Contrasting theories of White's illusion. Perception, 30, 1499-1501
    Blakeslee, B., Pasieka, W., & McCourt, M. E. (2005).
        Oriented multiscale spatial ﬁltering and contrast normalization:
        a parsimonious model of brightness induction in a continuum of stimuli
        including White, Howe and simultaneous brightness contrast.
        Vision Research, 45, 607-615.
    """

    height, width = 16.0, 16.0
    params = {
        "ppd": ppd,
        "frequency": 8.0 / width,
        "intensity_bars": (v3, v1),
        "intensity_target": v2,
        "target_indices_top": (6,),
        "target_indices_bottom": (11,),
        "target_center_offset": height / 10.0,
        "target_height": height / 5.0,
        "intensity_stripes": (v1, v3),
        "stripe_center_offset": height / 5.0,
        "stripe_height": height / 5.0,
        "period": "even",
    }

    stim = stimupy.whites.anderson(
        visual_size=(height, width),
        **params,
    )

    if pad:
        stim = pad_dict_to_visual_size(stim, VISEXTENT, ppd, pad_value=v2)

    experimental_data = {
        "effect_strength": 6.43,
    }

    params.update(
        visual_size=np.array(stim["img"].shape) / ppd,
        shape=stim["img"].shape,
        intensity_range=(v1, v3),
        experimental_data=experimental_data,
    )
    return {**stim, **params}


def WE_howe(ppd=PPD, pad=True):
    """Howe variation of White stimulus as shown in Robinson, Hammon, & de Sa (2007) Fig 1e.
    Grating size: 16x16 deg
    Grating frequency: 0.5 cpd
    Target size: 3.2x1 deg

    Parameters
    ----------
    ppd : int
        Resolution of stimulus in pixels per degree. (default: 32)
    pad : bool
        If True, include padding to 32x32 deg (default: True)

    Returns
    -------
    dict of str
        dict with the stimulus (key: "img") and target mask (key: "target_mask")
        and additional keys containing stimulus parameters

    References
    ----------
    Blakeslee, B., Pasieka, W., & McCourt, M. E. (2005).
        Oriented multiscale spatial ﬁltering and contrast normalization:
        a parsimonious model of brightness induction in a continuum
        of stimuli including White, Howe and simultaneous brightness contrast.
        Vision Research, 45, 607-615.
    Howe, P. D. L. (2001).
        A comment on the Anderson (1997), the Todorovic (1997),
        and the Ross and Pessoa (2000) explanations of White's eﬀect.
        Perception, 30, 1023-1026
    """

    height, width = 16.0, 16.0
    params = {
        "ppd": ppd,
        "frequency": 8.0 / width,
        "intensity_bars": (v3, v1),
        "intensity_target": v2,
        "target_indices_top": (6,),
        "target_indices_bottom": (11,),
        "target_center_offset": height / 5.0,
        "target_height": height / 5.0,
        "intensity_stripes": (v1, v3),
        "period": "even",
    }

    stim = stimupy.whites.howe(
        visual_size=(height, width),
        **params,
    )

    if pad:
        stim = pad_dict_to_visual_size(stim, VISEXTENT, ppd, pad_value=v2)

    experimental_data = {
        "effect_strength": 0.0,
    }

    params.update(
        visual_size=np.array(stim["img"].shape) / ppd,
        shape=stim["img"].shape,
        intensity_range=(v1, v3),
        experimental_data=experimental_data,
    )
    return {**stim, **params}


def WE_zigzag(ppd=PPD, pad=True):
    """Wedding cake stimulus as shown in Robinson, Hammon, & de Sa (2007) Fig 1f.

    Parameters
    ----------
    ppd : int
        Resolution of stimulus in pixels per degree. (default: 32)
    pad : bool
        If True, include padding to 32x32 deg (default: True)

    Returns
    -------
    dict of str
        dict with the stimulus (key: "img") and target mask (key: "target_mask")
        and additional keys containing stimulus parameters

    References
    ----------
    Clifford, C. W. G., & Spehar, B. (2003).
        Using colour to disambiguate contrast and assimilation in White's effect.
        Journal of Vision, 3, 294a.
        https://doi.org/10.1167/3.9.294
    """

    i1, i2 = 1, 2
    stim = stimupy.wedding_cakes.wedding_cake(
        visual_size=(16.0, 17.0),
        ppd=PPD,
        L_size=(4.0, 4, 1.0),
        target_height=2.0,
        target_indices1=((i1, -1), (i1, 0), (i1, 1), (i1, 2)),
        target_indices2=((i2, 0), (i2, 1), (i2, 2), (i2, 3)),
        intensity_bars=(1.0, 0.0),
        intensity_target=0.5,
    )

    if pad:
        stim = pad_dict_to_visual_size(stim, VISEXTENT, ppd, pad_value=v2)
    return stim


def WE_radial_thick_small(ppd=PPD, pad=True):
    """Radial White stimulus as shown in Robinson, Hammon, & de Sa (2007) Fig 1g.
    Circle size: 16x16 deg
    Number of segments: 14
    Target length: 4 deg

    Parameters
    ----------
    ppd : int
        Resolution of stimulus in pixels per degree. (default: 32)
    pad : bool
        If True, include padding to 32x32 deg (default: True)

    Returns
    -------
    dict of str
        dict with the stimulus (key: "img") and target mask (key: "target_mask")
        and additional keys containing stimulus parameters

    References
    ----------
    Anstis, S. (2003).
        White's eﬀect in brightness & color.
        Online Demonstration
    """

    n_segments = 14
    params = {
        "ppd": ppd,
        "n_segments": n_segments,
        "rotation": 90 - (90 / (n_segments / 2)),
        "target_indices": (1, 8),
        "target_width": 4.0,
        "target_center": 4.5,
        "intensity_segments": (v1, v3),
        "intensity_background": v2,
        "intensity_target": v2,
    }

    stim = stimupy.whites.angular(
        visual_size=(16, 16),
        **params,
    )

    if pad:
        stim = pad_dict_to_visual_size(stim, VISEXTENT, ppd, pad_value=v2)

    params.update(
        visual_size=np.array(stim["img"].shape) / ppd,
        shape=stim["img"].shape,
        intensity_range=(v1, v3),
    )
    return {**stim, **params}


def WE_radial_thick(ppd=PPD, pad=True):
    """Radial White stimulus as shown in Robinson, Hammon, & de Sa (2007) Fig 1h.
    Circle size: 24x24 deg
    Number of segments: 18
    Target length: 4 deg

    Parameters
    ----------
    ppd : int
        Resolution of stimulus in pixels per degree. (default: 32)
    pad : bool
        If True, include padding to 32x32 deg (default: True)

    Returns
    -------
    dict of str
        dict with the stimulus (key: "img") and target mask (key: "target_mask")
        and additional keys containing stimulus parameters

    References
    ----------
    Anstis, S. (2003).
        White's eﬀect in brightness & color.
        Online Demonstration
    """

    n_segments = 18
    params = {
        "ppd": ppd,
        "n_segments": n_segments,
        "rotation": 90 - (90 / (n_segments / 2)),
        "target_indices": (1, 10),
        "target_width": 4.0,
        "target_center": 6.0,
        "intensity_segments": (v1, v3),
        "intensity_background": v2,
        "intensity_target": v2,
    }

    stim = stimupy.whites.angular(
        visual_size=(24, 24),
        **params,
    )

    if pad:
        stim = pad_dict_to_visual_size(stim, VISEXTENT, ppd, pad_value=v2)

    params.update(
        visual_size=np.array(stim["img"].shape) / ppd,
        shape=stim["img"].shape,
        intensity_range=(v1, v3),
    )
    return {**stim, **params}


def WE_radial_thin_small(ppd=PPD, pad=True):
    """Radial White stimulus as shown in Robinson, Hammon, & de Sa (2007) Fig 1i.
    Circle size: 16x16 deg
    Number of segments: 26
    Target length: 2 deg

    Parameters
    ----------
    ppd : int
        Resolution of stimulus in pixels per degree. (default: 32)
    pad : bool
        If True, include padding to 32x32 deg (default: True)

    Returns
    -------
    dict of str
        dict with the stimulus (key: "img") and target mask (key: "target_mask")
        and additional keys containing stimulus parameters

    References
    ----------
    Anstis, S. (2003).
        White's eﬀect in brightness & color.
        Online Demonstration
    """

    n_segments = 26
    params = {
        "ppd": ppd,
        "n_segments": n_segments,
        "rotation": 90 - (90 / (n_segments / 2)),
        "target_indices": (1, 14),
        "target_width": 2.0,
        "target_center": 4.0,
        "intensity_segments": (v1, v3),
        "intensity_background": v2,
        "intensity_target": v2,
    }

    stim = stimupy.whites.angular(
        visual_size=(16, 16),
        **params,
    )

    if pad:
        stim = pad_dict_to_visual_size(stim, VISEXTENT, ppd, pad_value=v2)

    params.update(
        visual_size=np.array(stim["img"].shape) / ppd,
        shape=stim["img"].shape,
        intensity_range=(v1, v3),
    )
    return {**stim, **params}


def WE_radial_thin(ppd=PPD, pad=True):
    """Radial White stimulus as shown in Robinson, Hammon, & de Sa (2007) Fig 1j.
    Circle size: 24x24 deg
    Number of segments: 42
    Target length: 2 deg

    Parameters
    ----------
    ppd : int
        Resolution of stimulus in pixels per degree. (default: 32)
    pad : bool
        If True, include padding to 32x32 deg (default: True)

    Returns
    -------
    dict of str
        dict with the stimulus (key: "img") and target mask (key: "target_mask")
        and additional keys containing stimulus parameters

    References
    ----------
    Anstis, S. (2003).
        White's eﬀect in brightness & color.
        Online Demonstration
    """

    n_segments = 42
    params = {
        "ppd": ppd,
        "n_segments": n_segments,
        "rotation": 90 - (90 / (n_segments / 2)),
        "target_indices": (1, 22),
        "target_width": 2.0,
        "target_center": 6.0,
        "intensity_segments": (v1, v3),
        "intensity_background": v2,
        "intensity_target": v2,
    }

    stim = stimupy.whites.angular(
        visual_size=(24, 24),
        **params,
    )

    if pad:
        stim = pad_dict_to_visual_size(stim, VISEXTENT, ppd, pad_value=v2)

    params.update(
        visual_size=np.array(stim["img"].shape) / ppd,
        shape=stim["img"].shape,
        intensity_range=(v1, v3),
    )
    return {**stim, **params}


def WE_circular1(ppd=PPD, pad=True):
    """Ring (or circular White) stimulus as shown in Robinson, Hammon, & de Sa (2007) Fig 1k.
    Circle size: 16x16 deg
    Ring frequency: 0.5 cpd
    Ring width: 1 deg

    Parameters
    ----------
    ppd : int
        Resolution of stimulus in pixels per degree. (default: 32)
    pad : bool
        If True, include padding to 32x32 deg (default: True)

    Returns
    -------
    dict of str
        dict with the stimulus (key: "img") and target mask (key: "target_mask")
        and additional keys containing stimulus parameters

    References
    ----------
    Howe, P. D. L. (2005).
        White's eﬀect:
        removing the junctions but preserving the strength of the illusion.
        Perception, 34, 557-564.
    """

    height, width = 16.0, 16.0
    params = {
        "ppd": ppd,
        "frequency": 8.0 / height,
        "target_indices": 5,
        "intensity_background": v2,
        "intensity_target": v2,
        "clip": True,
    }

    stim1 = stimupy.whites.radial(
        visual_size=(height, width),
        intensity_rings=(v1, v3),
        **params,
    )

    stim2 = stimupy.whites.radial(
        visual_size=(height, width),
        intensity_rings=(v3, v1),
        **params,
    )
    stim = stack_dicts(stim1, stim2)

    if pad:
        stim = pad_dict_to_visual_size(stim, VISEXTENT, ppd, pad_value=v2)

    params.update(
        visual_size=np.array(stim["img"].shape) / ppd,
        shape=stim["img"].shape,
        intensity_range=(v1, v3),
    )
    return {**stim, **params}


def WE_circular05(ppd=PPD, pad=True):
    """Ring (or circular White) stimulus as shown in Robinson, Hammon, & de Sa (2007) Fig 1l.
    Circle size: 16x16 deg
    Ring frequency: 1 cpd
    Ring width: 0.5 deg

    Parameters
    ----------
    ppd : int
        Resolution of stimulus in pixels per degree. (default: 32)
    pad : bool
        If True, include padding to 32x32 deg (default: True)

    Returns
    -------
    dict of str
        dict with the stimulus (key: "img") and target mask (key: "target_mask")
        and additional keys containing stimulus parameters

    References
    ----------
    Howe, P. D. L. (2005).
        White's eﬀect:
        removing the junctions but preserving the strength of the illusion.
        Perception, 34, 557-564.
    """

    height, width = 16.0, 16.0
    params = {
        "ppd": ppd,
        "frequency": 16.0 / height,
        "target_indices": 11,
        "intensity_background": v2,
        "intensity_target": v2,
        "clip": True,
    }

    stim1 = stimupy.whites.radial(
        visual_size=(height, width),
        intensity_rings=(v1, v3),
        **params,
    )

    stim2 = stimupy.whites.radial(
        visual_size=(height, width),
        intensity_rings=(v3, v1),
        **params,
    )
    stim = stack_dicts(stim1, stim2)

    if pad:
        stim = pad_dict_to_visual_size(stim, VISEXTENT, ppd, pad_value=v2)

    params.update(
        visual_size=np.array(stim["img"].shape) / ppd,
        shape=stim["img"].shape,
        intensity_range=(v1, v3),
    )
    return {**stim, **params}


def WE_circular025(ppd=PPD, pad=True):
    """Ring (or circular White) stimulus as shown in Robinson, Hammon, & de Sa (2007) Fig 1m.
    Circle size: 16x16 deg
    Ring frequency: 2 cpd
    Ring width: 0.25 deg

    Parameters
    ----------
    ppd : int
        Resolution of stimulus in pixels per degree. (default: 32)
    pad : bool
        If True, include padding to 32x32 deg (default: True)

    Returns
    -------
    dict of str
        dict with the stimulus (key: "img") and target mask (key: "target_mask")
        and additional keys containing stimulus parameters

    References
    ----------
    Howe, P. D. L. (2005).
        White's eﬀect:
        removing the junctions but preserving the strength of the illusion.
        Perception, 34, 557-564.
    """

    height, width = 16.0, 16.0
    params = {
        "ppd": ppd,
        "frequency": 32.0 / height,
        "target_indices": 23,
        "intensity_background": v2,
        "intensity_target": v2,
        "clip": True,
    }

    stim1 = stimupy.whites.radial(
        visual_size=(height, width),
        intensity_rings=(v1, v3),
        **params,
    )

    stim2 = stimupy.whites.radial(
        visual_size=(height, width),
        intensity_rings=(v3, v1),
        **params,
    )
    stim = stack_dicts(stim1, stim2)

    if pad:
        stim = pad_dict_to_visual_size(stim, VISEXTENT, ppd, pad_value=v2)

    params.update(
        visual_size=np.array(stim["img"].shape) / ppd,
        shape=stim["img"].shape,
        intensity_range=(v1, v3),
    )
    return {**stim, **params}


def grating_induction(ppd=PPD, pad=True):
    """Grating induction as shown in Robinson, Hammon, & de Sa (2007) Fig 1n.
    Circle size: 12x16 deg
    Grating frequency: 0.25 cpd
    Target size: 1x16 deg

    Parameters
    ----------
    ppd : int
        Resolution of stimulus in pixels per degree. (default: 32)
    pad : bool
        If True, include padding to 32x32 deg (default: True)

    Returns
    -------
    dict of str
        dict with the stimulus (key: "img") and target mask (key: "target_mask")
        and additional keys containing stimulus parameters

    References
    ----------
    Blakeslee, B., & McCourt, M. E. (1999).
        A multiscale spatial ﬁltering account
        of the White eﬀect, simultaneous brightness contrast and grating induction.
        Vision Research, 39, 4361-4377.
    McCourt, M. E. (1982).
        A spatial frequency dependent grating-induction effect.
        Vision Research, 22, 119-134.
        https://doi.org/10.1016/0042-6989(82)90173-0
    """

    height, width = 12.0, 16.0
    params = {
        "ppd": ppd,
        "frequency": 4.0 / width,
        "target_width": 1.0,
        "intensities": (0.0, 1.0),
        "intensity_target": (0.5),
        "period": "ignore",
    }

    stim = stimupy.gratings.grating_induction(
        visual_size=(height, width),
        **params,
    )
    mask1 = np.where(stim["target_mask"] % 2, 1, 2)
    mask2 = np.where(stim["target_mask"] == 0, 0, 1)
    stim["target_mask"] = (mask1 * mask2).astype(int)

    if pad:
        stim = pad_dict_to_visual_size(stim, VISEXTENT, ppd, pad_value=v2)

    experimental_data = {
        "effect_strength": 6.23,
    }

    params.update(
        visual_size=np.array(stim["img"].shape) / ppd,
        shape=stim["img"].shape,
        intensity_range=(v1, v3),
        experimental_data=experimental_data,
    )
    return {**stim, **params}


def sbc_large(ppd=PPD, pad=True):
    """Simultaneous brightness contrast as shown in Robinson, Hammon, & de Sa (2007) Fig 1o.
    Stimulus size: 13x31 deg
    Target sizes: 3x3 deg

    Parameters
    ----------
    ppd : int
        Resolution of stimulus in pixels per degree. (default: 32)
    pad : bool
        If True, include padding to 32x32 deg (default: True)

    Returns
    -------
    dict of str
        dict with the stimulus (key: "img") and target mask (key: "target_mask")
        and additional keys containing stimulus parameters

    References
    ----------
    Blakeslee, B., & McCourt, M. E. (1999).
        A multiscale spatial ﬁltering account
        of the White eﬀect, simultaneous brightness contrast and grating induction.
        Vision Research, 39, 4361-4377.
    """

    stim = stimupy.sbcs.basic_two_sided(
        visual_size=(13.0, 31.0),
        ppd=ppd,
        target_size=3.0,
        intensity_target=v2,
        intensity_background=(0.0, 1.0),
    )

    if pad:
        stim = pad_dict_to_visual_size(stim, VISEXTENT, ppd, pad_value=v2)

    experimental_data = {
        "effect_strength": 11.35,
    }

    stim.update(
        visual_size=np.array(stim["img"].shape) / ppd,
        shape=stim["img"].shape,
        intensity_range=(v1, v3),
        experimental_data=experimental_data,
    )
    return stim


def sbc_small(ppd=PPD, pad=True):
    """Simultaneous brightness contrast as shown in Robinson, Hammon, & de Sa (2007) Fig 1p.
    Stimulus size: 13x31 deg
    Target sizes: 1x1 deg

    Parameters
    ----------
    ppd : int
        Resolution of stimulus in pixels per degree. (default: 32)
    pad : bool
        If True, include padding to 32x32 deg (default: True)

    Returns
    -------
    dict of str
        dict with the stimulus (key: "img") and target mask (key: "target_mask")
        and additional keys containing stimulus parameters

    References
    ----------
    Blakeslee, B., & McCourt, M. E. (1999).
        A multiscale spatial ﬁltering account
        of the White eﬀect, simultaneous brightness contrast and grating induction.
        Vision Research, 39, 4361-4377.
    """

    stim = stimupy.sbcs.basic_two_sided(
        visual_size=(13.0, 31.0),
        ppd=ppd,
        target_size=1.0,
        intensity_background=(0.0, 1.0),
        intensity_target=v2,
    )

    if pad:
        stim = pad_dict_to_visual_size(stim, VISEXTENT, ppd, pad_value=v2)

    experimental_data = {
        "effect_strength": 19.78,
    }

    stim.update(
        visual_size=np.array(stim["img"].shape) / ppd,
        shape=stim["img"].shape,
        intensity_range=(v1, v3),
        experimental_data=experimental_data,
    )
    return stim


def todorovic_equal(ppd=PPD, pad=True):
    """Todorovic stimulus - equal as shown in Robinson, Hammon, & de Sa (2007) Fig 1q.
    Stimulus size: 13x31 deg
    Target size: 8x8 deg

    Parameters
    ----------
    ppd : int
        Resolution of stimulus in pixels per degree. (default: 32)
    pad : bool
        If True, include padding to 32x32 deg (default: True)

    Returns
    -------
    dict of str
        dict with the stimulus (key: "img") and target mask (key: "target_mask")
        and additional keys containing stimulus parameters

    References
    ----------
    Blakeslee, B., & McCourt, M. E. (1999).
        A multiscale spatial ﬁltering account
        of the White eﬀect, simultaneous brightness contrast and grating induction.
        Vision Research, 39, 4361-4377.
    Pessoa, L., Baratoff, G., Neumann, H., & Todorovic, D. (1998).
        Lightness and junctions: variations on White's display.
        Investigative Ophthalmology and Visual Science (Supplement), 39, S159.
    """

    params = {
        "visual_size": (13.0, 15.5),
        "ppd": ppd,
        "cross_size": 8.0,
        "cross_thickness": 1.6,
        "intensity_target": v2,
    }

    stim1 = stimupy.todorovics.equal(
        intensity_background=1.0,
        intensity_covers=0.0,
        **params,
    )

    stim2 = stimupy.todorovics.equal(
        intensity_background=0.0,
        intensity_covers=1.0,
        **params,
    )
    stim = stack_dicts(stim1, stim2)

    if pad:
        stim = pad_dict_to_visual_size(stim, VISEXTENT, ppd, pad_value=v2)

    experimental_data = {
        "effect_strength": 2.2,
    }

    params.update(
        visual_size=np.array(stim["img"].shape) / ppd,
        shape=stim["img"].shape,
        intensity_range=(v1, v3),
        experimental_data=experimental_data,
    )
    return {**stim, **params}


def todorovic_in_large(ppd=PPD, pad=True):
    """Todorovic stimulus - in as shown in Robinson, Hammon, & de Sa (2007) Fig 1r.
    Stimulus size: 13x31 deg
    Target size: 5.3x5.3 deg

    Parameters
    ----------
    ppd : int
        Resolution of stimulus in pixels per degree. (default: 32)
    pad : bool
        If True, include padding to 32x32 deg (default: True)

    Returns
    -------
    dict of str
        dict with the stimulus (key: "img") and target mask (key: "target_mask")
        and additional keys containing stimulus parameters

    References
    ----------
    Blakeslee, B., & McCourt, M. E. (1999).
        A multiscale spatial ﬁltering account
        of the White eﬀect, simultaneous brightness contrast and grating induction.
        Vision Research, 39, 4361-4377.
    Todorovic, D. (1997).
        Lightness and junctions.
        Perception, 26, 379-395.
    """

    params = {
        "visual_size": (13.0, 15.5),
        "ppd": ppd,
        "target_size": 5.3125,
        "covers_size": 3.2,
        "covers_offset": 2.4,
        "intensity_target": v2,
    }

    stim1 = stimupy.todorovics.rectangle(
        intensity_background=1.0,
        intensity_covers=0.0,
        **params,
    )
    stim2 = stimupy.todorovics.rectangle(
        intensity_background=0.0,
        intensity_covers=1.0,
        **params,
    )
    stim = stack_dicts(stim1, stim2)

    if pad:
        stim = pad_dict_to_visual_size(stim, VISEXTENT, ppd, pad_value=v2)

    experimental_data = {
        "effect_strength": 2.4,
    }

    params.update(
        visual_size=np.array(stim["img"].shape) / ppd,
        shape=stim["img"].shape,
        intensity_range=(v1, v3),
        experimental_data=experimental_data,
    )
    return {**stim, **params}


def todorovic_in_small(ppd=PPD, pad=True):
    """Todorovic stimulus - in as shown in Robinson, Hammon, & de Sa (2007) Fig 1s.
    Stimulus size: 13x31 deg
    Target size: 3x3 deg

    Parameters
    ----------
    ppd : int
        Resolution of stimulus in pixels per degree. (default: 32)
    pad : bool
        If True, include padding to 32x32 deg (default: True)

    Returns
    -------
    dict of str
        dict with the stimulus (key: "img") and target mask (key: "target_mask")
        and additional keys containing stimulus parameters

    References
    ----------
    Blakeslee, B., & McCourt, M. E. (1999).
        A multiscale spatial ﬁltering account
        of the White eﬀect, simultaneous brightness contrast and grating induction.
        Vision Research, 39, 4361-4377.
    Todorovic, D. (1997).
        Lightness and junctions.
        Perception, 26, 379-395.
    """

    params = {
        "visual_size": (13.0, 15.5),
        "ppd": ppd,
        "cross_size": 3.0,
        "cross_thickness": 1.6,
        "covers_size": 3.2,
        "intensity_target": v2,
    }

    stim1 = stimupy.todorovics.cross(
        intensity_background=1.0,
        intensity_covers=0.0,
        **params,
    )
    stim2 = stimupy.todorovics.cross(
        intensity_background=0.0,
        intensity_covers=1.0,
        **params,
    )
    stim = stack_dicts(stim1, stim2)

    if pad:
        stim = pad_dict_to_visual_size(stim, VISEXTENT, ppd, pad_value=v2)

    experimental_data = {
        "effect_strength": 4.4,
    }

    params.update(
        visual_size=np.array(stim["img"].shape) / ppd,
        shape=stim["img"].shape,
        intensity_range=(v1, v3),
        experimental_data=experimental_data,
    )
    return {**stim, **params}


def todorovic_out(ppd=PPD, pad=True):
    """Todorovic stimulus - out as shown in Robinson, Hammon, & de Sa (2007) Fig 1t.
    Stimulus size: 13x31 deg
    Target size: 9.4x9.4 deg

    (note: in RHS2007, it says 8.7x8.7 deg, however that does not match with Fig 1t)

    Parameters
    ----------
    ppd : int
        Resolution of stimulus in pixels per degree. (default: 32)
    pad : bool
        If True, include padding to 32x32 deg (default: True)

    Returns
    -------
    dict of str
        dict with the stimulus (key: "img") and target mask (key: "target_mask")
        and additional keys containing stimulus parameters

    References
    ----------
    Blakeslee, B., & McCourt, M. E. (1999).
        A multiscale spatial ﬁltering account
        of the White eﬀect, simultaneous brightness contrast and grating induction.
        Vision Research, 39, 4361-4377.
    Pessoa, L., Baratoff, G., Neumann, H., & Todorovic, D. (1998).
        Lightness and junctions: variations on White's display.
        Investigative Ophthalmology and Visual Science (Supplement), 39, S159.
    """

    params = {
        "visual_size": (13.0, 15.5),
        "ppd": ppd,
        "cross_size": 9.2,
        "cross_thickness": 1.6,
        "covers_size": 3.2,
        "intensity_target": v2,
    }

    stim1 = stimupy.todorovics.cross(
        intensity_background=1.0,
        intensity_covers=0.0,
        **params,
    )
    stim2 = stimupy.todorovics.cross(
        intensity_background=0.0,
        intensity_covers=1.0,
        **params,
    )
    stim = stack_dicts(stim1, stim2)

    if pad:
        stim = pad_dict_to_visual_size(stim, VISEXTENT, ppd, pad_value=v2)

    experimental_data = {
        "effect_strength": 1.53,
    }

    params.update(
        visual_size=np.array(stim["img"].shape) / ppd,
        shape=stim["img"].shape,
        intensity_range=(v1, v3),
        experimental_data=experimental_data,
    )
    return {**stim, **params}


def checkerboard_016(ppd=PPD, pad=True):
    """Checkerboard stimulus as shown in Robinson, Hammon, & de Sa (2007) Fig 1u.
    Stimulus size: 6.25x15.9375 deg
    Target size: 0.156x0.156 deg

    Parameters
    ----------
    ppd : int
        Resolution of stimulus in pixels per degree. (default: 32)
    pad : bool
        If True, include padding to 32x32 deg (default: True)

    Returns
    -------
    dict of str
        dict with the stimulus (key: "img") and target mask (key: "target_mask")
        and additional keys containing stimulus parameters

    References
    ----------
    Blakeslee, B., & McCourt, M. E. (2004).
        A uniﬁed theory of brightness contrast and assimilation
        incorporating oriented multi-scale spatial ﬁltering and contrast normalization.
        Vision Research, 44, 2483-2503.
    DeValois, R. L., & DeValois, K. K. (1988).
        Spatial vision.
        New York: Oxford University Press.
    """

    nchecks_height, nchecks_width = 40, 102
    target_row = nchecks_height // 2
    params = {
        "ppd": ppd,
        "board_shape": (nchecks_height, nchecks_width),
        "check_visual_size": (5 / 32, 5 / 32),
        "target_indices": ((target_row, 16), (target_row, 85)),
        "extend_targets": False,
        "intensity_checks": (v3, v1),
        "intensity_target": v2,
    }
    stim = stimupy.checkerboards.checkerboard(**params)

    if pad:
        stim = pad_dict_to_visual_size(stim, VISEXTENT, ppd, pad_value=v2)

    experimental_data = {
        "effect_strength": 7.46,
    }

    params.update(
        visual_size=np.array(stim["img"].shape) / ppd,
        shape=stim["img"].shape,
        intensity_range=(v1, v3),
        experimental_data=experimental_data,
    )
    return {**stim, **params}


def checkerboard_094(ppd=PPD, pad=True):
    """Checkerboard stimulus as shown in Robinson, Hammon, & de Sa (2007) Fig 1v.
    Stimulus size: 6.5625x23.4375 deg
    Target size: 0.938x0.938 deg

    Parameters
    ----------
    ppd : int
        Resolution of stimulus in pixels per degree. (default: 32)
    pad : bool
        If True, include padding to 32x32 deg (default: True)

    Returns
    -------
    dict of str
        dict with the stimulus (key: "img") and target mask (key: "target_mask")
        and additional keys containing stimulus parameters

    References
    ----------
    Blakeslee, B., & McCourt, M. E. (2004).
        A uniﬁed theory of brightness contrast and assimilation
        incorporating oriented multi-scale spatial ﬁltering and contrast normalization.
        Vision Research, 44, 2483-2503.
    DeValois, R. L., & DeValois, K. K. (1988).
        Spatial vision.
        New York: Oxford University Press.
    """

    nchecks_height, nchecks_width = 7, 25
    target_row = nchecks_height // 2
    params = {
        "ppd": ppd,
        "board_shape": (nchecks_height, nchecks_width),
        "check_visual_size": (30 / 32, 30 / 32),
        "target_indices": ((target_row, 6), (target_row, 17)),
        "extend_targets": False,
        "intensity_checks": (v1, v3),
        "intensity_target": v2,
    }
    stim = stimupy.checkerboards.checkerboard(**params)

    if pad:
        stim = pad_dict_to_visual_size(stim, VISEXTENT, ppd, pad_value=v2)

    experimental_data = {
        "effect_strength": 2.84,
    }

    params.update(
        visual_size=np.array(stim["img"].shape) / ppd,
        shape=stim["img"].shape,
        intensity_range=(v1, v3),
        experimental_data=experimental_data,
    )
    return {**stim, **params}


def checkerboard_21(ppd=PPD, pad=True):
    """Checkerboard stimulus as shown in Robinson, Hammon, & de Sa (2007) Fig 1w.
    Stimulus size: 6.28125x20.9375 deg
    Target size: 2.09x2.09 deg

    Parameters
    ----------
    ppd : int
        Resolution of stimulus in pixels per degree. (default: 32)
    pad : bool
        If True, include padding to 32x32 deg (default: True)

    Returns
    -------
    dict of str
        dict with the stimulus (key: "img") and target mask (key: "target_mask")
        and additional keys containing stimulus parameters

    References
    ----------
    Blakeslee, B., & McCourt, M. E. (2004).
        A uniﬁed theory of brightness contrast and assimilation
        incorporating oriented multi-scale spatial ﬁltering and contrast normalization.
        Vision Research, 44, 2483-2503.
    DeValois, R. L., & DeValois, K. K. (1988).
        Spatial vision.
        New York: Oxford University Press.
    """

    nchecks_height, nchecks_width = 3, 10
    target_row = nchecks_height // 2
    params = {
        "ppd": ppd,
        "board_shape": (nchecks_height, nchecks_width),
        "check_visual_size": (67 / 32, 67 / 32),
        "target_indices": ((target_row, 2), (target_row, 7)),
        "extend_targets": False,
        "intensity_checks": (v1, v3),
        "intensity_target": v2,
    }
    stim = stimupy.checkerboards.checkerboard(**params)

    if pad:
        stim = pad_dict_to_visual_size(stim, VISEXTENT, ppd, pad_value=v2)

    experimental_data = {
        "effect_strength": 5.67,
    }

    params.update(
        visual_size=np.array(stim["img"].shape) / ppd,
        shape=stim["img"].shape,
        intensity_range=(v1, v3),
        experimental_data=experimental_data,
    )
    return {**stim, **params}


def corrugated_mondrian(ppd=PPD, pad=True):
    """Corrugated Mondrians as shown in Robinson, Hammon, & de Sa (2007) Fig 1x.

    Parameters
    ----------
    ppd : int
        Resolution of stimulus in pixels per degree. (default: 32)
    pad : bool
        If True, include padding to 32x32 deg (default: True)

    Returns
    -------
    dict of str
        dict with the stimulus (key: "img") and target mask (key: "target_mask")
        and additional keys containing stimulus parameters

    References
    ----------
    Blakeslee, B., & McCourt, M. E. (2001).
        A multiscale spatial filtering account
        of the Wertheimer-Benary effect and the corrugated Mondrian.
        Vision Research, 41, 2487-2502.
    Adelson, E. H. (1993).
        Perceptual organization and the judgment of brightness.
        Science, 262, 2042-2044.
        https://doi.org/10.1126/science.8266102
    """

    v1, v2, v3, v4 = 0.0, 0.4, 0.75, 1.0
    values = (
        (v3, v2, v3, v2, v3),
        (v4, v3, v2, v3, v4),
        (v3, v2, v3, v2, v3),
        (v2, v1, v2, v1, v2),
        (v3, v2, v3, v2, v3),
    )
    params = {
        "visual_size": (5 * 2, 5 * 2 + 1),
        "ppd": ppd,
        "depths": (0.0, -1.0, 0.0, 1.0, 0.0),
        "intensities": values,
        "target_indices": ((1, 2), (3, 2)),
        "intensity_background": 0.5,
    }

    stim = stimupy.mondrians.corrugated_mondrian(**params)

    if pad:
        stim = pad_dict_to_visual_size(stim, VISEXTENT, ppd, pad_value=0.5)

    experimental_data = {
        "effect_strength": 10.85,
    }

    params.update(
        visual_size=np.array(stim["img"].shape) / ppd,
        shape=stim["img"].shape,
        intensity_range=(v1, v4),
        experimental_data=experimental_data,
    )
    return {**stim, **params}


def benary_cross(ppd=PPD, pad=True):
    """Benarys cross as shown in Robinson, Hammon, & de Sa (2007) Fig 1y.

    Parameters
    ----------
    ppd : int
        Resolution of stimulus in pixels per degree. (default: 32)
    pad : bool
        If True, include padding to 32x32 deg (default: True)

    Returns
    -------
    dict of str
        dict with the stimulus (key: "img") and target mask (key: "target_mask")
        and additional keys containing stimulus parameters

    References
    ----------
    Blakeslee, B., & McCourt, M. E. (2001).
        A multiscale spatial filtering account
        of the Wertheimer-Benary effect and the corrugated Mondrian.
        Vision Research, 41, 2487-2502.
    Benary, W. (1924).
        Beobachtungen zu einem Experiment über Helligkeitskontrast.
        Psychologische Forschung, 5, 131-142.
        https://doi.org/10.1007/BF00402398
    """

    params = {
        "ppd": ppd,
        "cross_thickness": 4.0,
        "target_size": 2.5,
        "intensity_background": v3,
        "intensity_cross": v1,
        "intensity_target": v2,
    }

    stim = stimupy.benarys.cross_triangles(
        visual_size=(13, 23),
        **params,
    )
    stim = flip_dict(stim)

    # Switch target indices
    mask1 = np.where(stim["target_mask"] == 1, 1, 0)
    mask2 = np.where(stim["target_mask"] == 2, 1, 0)
    stim["target_mask"] = (mask1 * 2 + mask2).astype(int)

    if pad:
        stim = pad_dict_by_visual_size(stim, ((0.0, 0.0), (4.0, 4.0)), ppd, pad_value=v3)
        stim = pad_dict_to_visual_size(stim, VISEXTENT, ppd, pad_value=v2)

    experimental_data = {
        "effect_strength": 9.2,
    }

    params.update(
        visual_size=np.array(stim["img"].shape) / ppd,
        shape=stim["img"].shape,
        intensity_range=(v1, v3),
        experimental_data=experimental_data,
    )
    return {**stim, **params}


def todorovic_benary1_2(ppd=PPD, pad=True):
    """Todorovic variation of Benarys cross as shown in Robinson, Hammon, & de Sa (2007) Fig 1z.
    Only contains the first two targets (1-2).

    Parameters
    ----------
    ppd : int
        Resolution of stimulus in pixels per degree. (default: 32)
    pad : bool
        If True, include padding to 32x32 deg (default: True)

    Returns
    -------
    dict of str
        dict with the stimulus (key: "img") and target mask (key: "target_mask")
        and additional keys containing stimulus parameters

    References
    ----------
    Blakeslee, B., & McCourt, M. E. (2001).
        A multiscale spatial filtering account
        of the Wertheimer-Benary effect and the corrugated Mondrian.
        Vision Research, 41, 2487-2502.
    Todorovic, D. (1997).
        Lightness and junctions.
        Perception, 26, 379-395.
    """

    # Generate full stim
    stim = todorovic_benary1_2_3_4(ppd=ppd, pad=pad)

    # Remove unused targets
    stim["img"] = np.where(stim["target_mask"] == 3, v3, stim["img"])
    stim["img"] = np.where(stim["target_mask"] == 4, v1, stim["img"])
    stim["target_mask"] = np.where(stim["target_mask"] > 2, 0, stim["target_mask"])
    stim["target_type"] = stim["target_type"][:2]
    stim["target_rotation"] = stim["target_rotation"][:2]
    stim["target_x"] = stim["target_x"][:2]
    stim["target_y"] = stim["target_y"][:2]

    # Add experimental data
    stim["effect_strength"] = 11.96

    return stim


def todorovic_benary3_4(ppd=PPD, pad=True):
    """Todorovic variation of Benarys cross as shown in Robinson, Hammon, & de Sa (2007) Fig 1z.
    Only contains the last two targets (3-4).

    Parameters
    ----------
    ppd : int
        Resolution of stimulus in pixels per degree. (default: 32)
    pad : bool
        If True, include padding to 32x32 deg (default: True)

    Returns
    -------
    dict of str
        dict with the stimulus (key: "img") and target mask (key: "target_mask")
        and additional keys containing stimulus parameters

    References
    ----------
    Blakeslee, B., & McCourt, M. E. (2001).
        A multiscale spatial filtering account
        of the Wertheimer-Benary effect and the corrugated Mondrian.
        Vision Research, 41, 2487-2502.
    Todorovic, D. (1997).
        Lightness and junctions.
        Perception, 26, 379-395.
    """

    # Generate full stim
    stim = todorovic_benary1_2_3_4(ppd=ppd, pad=pad)

    # Remove unused targets
    stim["img"] = np.where(stim["target_mask"] == 1, v3, stim["img"])
    stim["img"] = np.where(stim["target_mask"] == 2, v1, stim["img"])
    stim["target_mask"] = np.where(stim["target_mask"] < 3, 0, stim["target_mask"])
    stim["target_mask"] = np.where(
        stim["target_mask"], stim["target_mask"] - 2, stim["target_mask"]
    )
    stim["target_type"] = stim["target_type"][2:]
    stim["target_rotation"] = stim["target_rotation"][2:]
    stim["target_x"] = stim["target_x"][2:]
    stim["target_y"] = stim["target_y"][2:]

    # Add experimental data
    stim["effect_strength"] = 9.55

    return stim


def todorovic_benary1_2_3_4(ppd=PPD, pad=True):
    """Todorovic variation of Benarys cross as shown in Robinson, Hammon, & de Sa (2007) Fig 1z.
    Contains all four targets (1-2-3-4).

    Parameters
    ----------
    ppd : int
        Resolution of stimulus in pixels per degree. (default: 32)
    pad : bool
        If True, include padding to 32x32 deg (default: True)

    Returns
    -------
    dict of str
        dict with the stimulus (key: "img") and target mask (key: "target_mask")
        and additional keys containing stimulus parameters

    References
    ----------
    Blakeslee, B., & McCourt, M. E. (2001).
        A multiscale spatial filtering account
        of the Wertheimer-Benary effect and the corrugated Mondrian.
        Vision Research, 41, 2487-2502.
    Todorovic, D. (1997).
        Lightness and junctions.
        Perception, 26, 379-395.
    """

    params = {
        "ppd": ppd,
        "L_width": 2.5,
        "target_size": 2.5,
        "target_type": ("t", "t", "t", "t"),
        "target_rotation": (0.0, 45.0, 225.0, 180.0),
        "target_x": (2.5, 9.5, 18.0, 26.0),
        "target_y": (4.0, 6.5, 6.5 - np.sqrt(12.5) / 2.0 + 1 / ppd, 6.5),
        "intensity_background": v3,
        "intensity_cross": v1,
        "intensity_target": v2,
    }

    stim = stimupy.benarys.todorovic_generalized(
        visual_size=(13.0, 31.0),
        **params,
    )

    if pad:
        stim = pad_dict_to_visual_size(stim, VISEXTENT, ppd, pad_value=v2)

    params.update(
        visual_size=np.array(stim["img"].shape) / ppd,
        shape=stim["img"].shape,
        intensity_range=(v1, v3),
    )
    return {**stim, **params}


def bullseye_thin(ppd=PPD, pad=True):
    """Bullseye stimulus as shown in Robinson, Hammon, & de Sa (2007) Fig 1aa.
    Target size: 0.608x0.608 deg
    Ring widths: 0.122 deg

    Parameters
    ----------
    ppd : int
        Resolution of stimulus in pixels per degree. (default: 32)
    pad : bool
        If True, include padding to 32x32 deg (default: True)

    Returns
    -------
    dict of str
        dict with the stimulus (key: "img") and target mask (key: "target_mask")
        and additional keys containing stimulus parameters

    References
    ----------
    Bindman, D., & Chubb, C. (2004).
        Brightness assimilation in bullseye displays.
        Vision Research, 44, 309-319.
        https://doi.org/10.1016/S0042-6989(03)00430-9
    """

    radii = np.array([0.304, 0.426, 0.548, 0.670, 0.792])

    params = {
        "visual_size": 0.792 * 2,
        "ppd": ppd,
        "radii": radii,
        "intensity_target": v2,
        "origin": "mean",
    }

    stim1 = stimupy.bullseyes.rectangular_generalized(
        **params,
        intensity_frames=(v1, v3),
    )
    stim2 = stimupy.bullseyes.rectangular_generalized(
        **params,
        intensity_frames=(v3, v1),
    )

    # Individual padding
    if pad:
        stim1 = pad_dict_to_visual_size(stim1, np.array(VISEXTENT) / 2, ppd, pad_value=v2)
        stim2 = pad_dict_to_visual_size(stim2, np.array(VISEXTENT) / 2, ppd, pad_value=v2)
    stim = stack_dicts(stim1, stim2)

    if pad:
        stim = pad_dict_to_visual_size(stim, VISEXTENT, ppd, pad_value=v2)

    params.update(
        visual_size=np.array(stim["img"].shape) / ppd,
        shape=stim["img"].shape,
        intensity_range=(v1, v3),
    )
    return {**stim, **params}


def bullseye_thick(ppd=PPD, pad=True):
    """Bullseye stimulus as shown in Robinson, Hammon, & de Sa (2007) Fig 1bb.
    Target size: 0.608x0.608 deg
    Ring widths: 0.243 deg

    Parameters
    ----------
    ppd : int
        Resolution of stimulus in pixels per degree. (default: 32)
    pad : bool
        If True, include padding to 32x32 deg (default: True)

    Returns
    -------
    dict of str
        dict with the stimulus (key: "img") and target mask (key: "target_mask")
        and additional keys containing stimulus parameters

    References
    ----------
    Bindman, D., & Chubb, C. (2004).
        Brightness assimilation in bullseye displays.
        Vision Research, 44, 309-319.
        https://doi.org/10.1016/S0042-6989(03)00430-9
    """
    radii = np.array([0.304, 0.547, 0.790, 1.033, 1.276])

    params = {
        "visual_size": 1.276 * 2,
        "ppd": ppd,
        "radii": radii,
        "intensity_target": v2,
        "origin": "mean",
    }

    stim1 = stimupy.bullseyes.rectangular_generalized(
        **params,
        intensity_frames=(v1, v3),
    )
    stim2 = stimupy.bullseyes.rectangular_generalized(
        **params,
        intensity_frames=(v3, v1),
    )

    # Individual padding
    if pad:
        stim1 = pad_dict_to_visual_size(stim1, np.array(VISEXTENT) / 2, ppd, pad_value=v2)
        stim2 = pad_dict_to_visual_size(stim2, np.array(VISEXTENT) / 2, ppd, pad_value=v2)
    stim = stack_dicts(stim1, stim2)

    if pad:
        stim = pad_dict_to_visual_size(stim, VISEXTENT, ppd, pad_value=v2)

    params.update(
        visual_size=np.array(stim["img"].shape) / ppd,
        shape=stim["img"].shape,
        intensity_range=(v1, v3),
    )
    return {**stim, **params}


if __name__ == "__main__":
    from stimupy.utils import plot_stimuli

    stims = gen_all(pad=True, skip=True)
    plot_stimuli(stims, mask=True)



from stimupy.stimuli import gabors as gabors_stim
from stimupy.stimuli import waves

__all__ = [
    "gabors",
    "sine_waves",
    "square_waves",
]


def add_waves(wave_dict1, wave_dict2, weight1=1, weight2=1):
    """
    Create plaid-like stimulus by adding two waves

    Parameters
    ----------
    wave_dict1 : dict
        dictionary which contains the first wave-array (key: "img"), as well as
        keys "shape" and "ppd"
    wave_dict2 : dict
        dictionary which contains the second wave-array (key: "img"), as well as
        keys "shape" and "ppd"
    weight1 : float, optional
        Factor with which the first wave is multiplied. The default is 1.
    weight2 : float, optional
        Factor with which the second wave is multiplied. The default is 1.

    Returns
    -------
    wave_dict1 : dict
        dictionary with plaid-like stimulus and additional keys if specified.

    """
    if wave_dict1["shape"] != wave_dict2["shape"]:
        raise ValueError(
            f"Waves have different shapes; 1: {wave_dict1['shape']}, 2: {wave_dict2['shape']}"
        )
    if wave_dict1["ppd"] != wave_dict2["ppd"]:
        raise ValueError(
            "Waves have different ppds; 1: {wave_dict1['ppd']}, 2: {wave_dict2['ppd']}"
        )

    img = weight1 * wave_dict1["img"] + weight2 * wave_dict2["img"]
    img = img / (weight1 + weight2)

    # Update parameters
    wave_dict1["img"] = img
    try:
        wave_dict1["grating_mask2"] = wave_dict2["grating_mask"]
        wave_dict1["frequency2"] = wave_dict2["frequency"]
        wave_dict1["phase_width2"] = wave_dict2["phase_width"]
        wave_dict1["n_phases2"] = wave_dict2["n_phases"]
    except Exception:
        pass
    return wave_dict1


def gabors(
    gabor_parameters1,
    gabor_parameters2,
    weight1=1,
    weight2=1,
):
    """Draw plaid consisting of two gabors

    Parameters
    ----------
    gabor_parameters1 : dict
        kwargs to generate first Gabor
    gabor_parameters2 : dict
        kwargs to generate second Gabor
    weight1 : float
        weight of first Gabor (default: 1)
    weight2 : float
        weight of second Gabor (default: 1)

    Returns
    -------
    dict[str, Any]
        dict with the stimulus (key: "img"),
        mask with integer index for each phase (key: "grating_mask"),
        and additional keys containing stimulus parameters
    """

    # Create sine-wave gratings
    grating1 = gabors_stim.gabor(**gabor_parameters1)
    grating2 = gabors_stim.gabor(**gabor_parameters2)
    plaid = add_waves(grating1, grating2, weight1, weight2)

    out = {
        "img": plaid["img"],
        "grating_mask1": plaid["grating_mask"],
        "grating_mask2": plaid["grating_mask2"],
        "gabor_parameters1": gabor_parameters1,
        "gabor_parameters2": gabor_parameters2,
        "weight1": weight1,
        "weight2": weight2,
        "visual_size": plaid["visual_size"],
        "shape": plaid["shape"],
        "ppd": plaid["ppd"],
    }
    return out


def sine_waves(
    grating_parameters1,
    grating_parameters2,
    weight1=1,
    weight2=1,
):
    """Draw plaid consisting of two sine-wave gratings

    Parameters
    ----------
    grating_parameters1 : dict
        kwargs to generate first sine-wave grating
    grating_parameters2 : dict
        kwargs to generate second sine-wave grating
    weight1 : float
        weight of first grating (default: 1)
    weight2 : float
        weight of second grating (default: 1)

    Returns
    -------
    dict[str, Any]
        dict with the stimulus (key: "img"),
        mask with integer index for each phase (key: "grating_mask"),
        and additional keys containing stimulus parameters
    """

    # Create sine-wave gratings
    grating1 = waves.sine_linear(**grating_parameters1)
    grating2 = waves.sine_linear(**grating_parameters2)
    plaid = add_waves(grating1, grating2, weight1, weight2)

    out = {
        "img": plaid["img"],
        "grating_mask1": plaid["grating_mask"],
        "grating_mask2": plaid["grating_mask2"],
        "grating_parameters1": grating_parameters1,
        "grating_parameters2": grating_parameters2,
        "weight1": weight1,
        "weight2": weight2,
        "visual_size": plaid["visual_size"],
        "shape": plaid["shape"],
        "ppd": plaid["ppd"],
    }
    return out


def square_waves(
    grating_parameters1,
    grating_parameters2,
    weight1=1,
    weight2=1,
):
    """Draw plaid consisting of two square-wave gratings

    Parameters
    ----------
    grating_parameters1 : dict
        kwargs to generate first sine-wave grating
    grating_parameters2 : dict
        kwargs to generate second sine-wave grating
    weight1 : float
        weight of first grating (default: 1)
    weight2 : float
        weight of second grating (default: 1)

    Returns
    -------
    dict[str, Any]
        dict with the stimulus (key: "img"),
        mask with integer index for each phase (key: "grating_mask"),
        and additional keys containing stimulus parameters
    """

    # Create sine-wave gratings
    grating1 = waves.square_linear(**grating_parameters1)
    grating2 = waves.square_linear(**grating_parameters2)
    plaid = add_waves(grating1, grating2, weight1, weight2)

    out = {
        "img": plaid["img"],
        "grating_mask1": plaid["grating_mask"],
        "grating_mask2": plaid["grating_mask2"],
        "grating_parameters1": grating_parameters1,
        "grating_parameters2": grating_parameters2,
        "weight1": weight1,
        "weight2": weight2,
        "visual_size": plaid["visual_size"],
        "shape": plaid["shape"],
        "ppd": plaid["ppd"],
    }
    return out


def overview(**kwargs):
    """Generate example stimuli from this module

    Returns
    -------
    stims : dict
        dict with all stimuli containing individual stimulus dicts.
    """
    default_params = {
        "visual_size": 15,
        "ppd": 10,
        "origin": "center",
        "phase_shift": 30,
    }
    default_params.update(kwargs)

    grating1 = {
        **default_params,
        "bar_width": 1,
        "period": "ignore",
        "rotation": 0,
        "round_phase_width": False,
    }

    grating2 = {
        **default_params,
        "bar_width": 0.5,
        "period": "ignore",
        "rotation": 90,
        "round_phase_width": False,
    }

    # fmt: off
    stimuli = {
        "plaid_gabors": gabors({**grating1, "sigma": 3}, {**grating2, "sigma": 3}),
        "plaid_sine_waves": sine_waves(grating1, grating2),
        "plaid_square_waves": square_waves(grating1, grating2),
    }
    # fmt: on

    return stimuli


if __name__ == "__main__":
    from stimupy.utils import plot_stimuli

    stims = overview()
    plot_stimuli(stims, mask=False, save=None)
